import datetime
import json
from typing import cast
from unittest import mock

from alliance_platform.codegen.printer import TypescriptPrinter
from alliance_platform.frontend.bundler.base import HtmlGenerationTarget
from alliance_platform.frontend.bundler.context import BundlerAssetContext
from alliance_platform.frontend.bundler.ssr import SSRJsonEncoder
from alliance_platform.frontend.bundler.ssr import SSRSerializerContext
from alliance_platform.frontend.bundler.vanilla_extract import resolve_vanilla_extract_cache_names
from alliance_platform.frontend.templatetags.react import ComponentNode
from alliance_platform.frontend.templatetags.react import ComponentProps
from alliance_platform.frontend.templatetags.react import ComponentSourceCodeGenerator
from django.conf import settings
from django.template import Context
from django.template import Template
from django.test import SimpleTestCase
from django.test import override_settings
from django.utils.functional import SimpleLazyObject
from django.utils.functional import lazy
from django.utils.safestring import mark_safe
from django.utils.timezone import make_aware

from .test_utils import override_ap_frontend_settings
from .test_utils.bundler import TestViteBundler
from .test_utils.bundler import bundler_kwargs
from .test_utils.bundler import bypass_frontend_asset_registry
from .test_utils.bundler import fixtures_dir
from .test_utils.bundler import format_code
from .test_utils.bundler import run_prettier

inline_css_prod = {
    fixtures_dir / "build_test/assets/Button-abc123.css": ".prod_button { color: red; }",
}


def mock_read_text(path):
    if path in inline_css_prod:
        return inline_css_prod[path]
    raise NotImplementedError(f"Mock read_text not implemented for {path}")


def mock_create_mapping(self):
    # Bypass the requests to the dev server VanillaExtractMapping does. They still work
    # but will log warnings because dev server won't be running
    self._load_mapping()


@override_settings(STATIC_URL="/static/")
@mock.patch(
    "alliance_platform.frontend.bundler.vanilla_extract.VanillaExtractClassMapping._create_mapping",
    mock_create_mapping,
)
class TestBundlerTemplateTags(SimpleTestCase):
    def setUp(self) -> None:
        self.test_production_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="production",
        )

        self.test_development_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="development",
        )
        self.dev_url = self.test_development_bundler.dev_server_url

    def test_bundler_url(self):
        for bundler_name, expected in [
            ("test_development_bundler", f"{self.dev_url}components/Button.css"),
            ("test_production_bundler", "/static/assets/Button-abc123.css"),
        ]:
            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry):
                with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                    tpl = Template("{% load bundler %}{% bundler_url 'components/Button.css' %}")
                    actual = tpl.render(Context())
                    self.assertEqual(
                        expected,
                        actual,
                    )

    def test_bundler_embed(self):
        for bundler_name, expected in [
            (
                "test_development_bundler",
                f'<script src="{self.dev_url}components/Button.tsx" type="module"></script>',
            ),
            (
                "test_production_bundler",
                '<script src="/static/assets/Button-def456.js" type="module"></script>\n'
                '<link rel="stylesheet" href="/static/assets/Button-abc123.css">',
            ),
        ]:
            with self.subTest(bundler_name=bundler_name):
                with BundlerAssetContext(
                    frontend_asset_registry=bypass_frontend_asset_registry
                ) as asset_context:
                    with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                        tpl = Template(
                            "{% load bundler %}{% bundler_embed 'components/Button.tsx' inline=True %}"
                        )
                        actual = asset_context.post_process(tpl.render(Context()))
                        self.assertEqual(
                            expected,
                            actual,
                        )

    def test_bundler_embed_collected_assets_check(self):
        """Check that bundler_embed_collected_assets exists if it's required by other tags"""
        with override_ap_frontend_settings(
            BUNDLER=self.test_development_bundler,
        ):
            with self.assertRaisesMessage(
                ValueError, "BundlerAssetContext.post_process() was not called but is required"
            ):
                with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry):
                    Template("{% load bundler %}{% bundler_embed 'components/Button.tsx' %}").render(
                        Context()
                    )

            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry) as asset_context:
                Template("{% load bundler %}{% bundler_embed 'components/Button.tsx' %}").render(Context())
                asset_context.post_process(
                    Template("{% load bundler %}{% bundler_embed_collected_assets %}").render(Context())
                )

    def test_bundler_embed_collected_assets_no_duplicate(self):
        """Check that bundler_embed_collected_assets exists only once"""
        with self.assertRaisesMessage(ValueError, "Duplicate {% bundler_embed_collected_assets %}"):
            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry):
                Template("{% load bundler %}{% bundler_embed_collected_assets %}").render(Context())
                Template("{% load bundler %}{% bundler_embed_collected_assets %}").render(Context())

    def test_bundler_embed_collected_assets(self):
        for bundler, expected in [
            (
                self.test_development_bundler,
                f'<script src="{self.dev_url}components/Button.tsx" type="module"></script>',
            ),
            (
                self.test_production_bundler,
                '<script src="/static/assets/Button-def456.js" type="module"></script>\n'
                '<link rel="stylesheet" href="/static/assets/Button-abc123.css">',
            ),
        ]:
            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry) as asset_context:
                with override_ap_frontend_settings(BUNDLER=bundler):
                    context = Context()
                    Template("{% load bundler %}{% bundler_embed 'components/Button.tsx' %}").render(context)
                    tpl = Template("{% load bundler %}{% bundler_embed_collected_assets %}")
                    actual = asset_context.post_process(tpl.render(context))
                    self.assertEqual(
                        expected,
                        actual,
                    )

    def test_bundler_embed_collected_assets_order(self):
        """Test ordering is preserved as it might be important"""
        for bundler_name, expected in [
            (
                "test_development_bundler",
                # yes, in dev css is loaded via <script>
                f'<script src="{self.dev_url}styles/normalize.css" type="module"></script>\n'
                f'<script src="{self.dev_url}components/Button.tsx" type="module"></script>\n'
                f'<script src="{self.dev_url}{resolve_vanilla_extract_cache_names(self.test_development_bundler, "login.css.ts")[1].relative_to(settings.PROJECT_DIR)}" type="module" blocking="render"></script>',
            ),
            (
                "test_production_bundler",
                '<link rel="stylesheet" href="/static/assets/normalize-x1.css">\n'
                '<script src="/static/assets/Button-def456.js" type="module"></script>\n'
                '<link rel="stylesheet" href="/static/assets/Button-abc123.css">\n'
                # While this file is generated ViteCssEmbed specifically excludes it as
                # it knows it's not needed
                # '<script src="/static/assets/login.css-669fd56c.js" type="module"></script>\n'
                '<link rel="stylesheet" href="/static/assets/login.css-a9690449.css">',
            ),
        ]:
            with self.subTest(bundler_name=bundler_name):
                with BundlerAssetContext(
                    frontend_asset_registry=bypass_frontend_asset_registry
                ) as asset_context:
                    with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                        context = Context()
                        Template(
                            """
                            {% load bundler %}
                            {% bundler_embed 'styles/normalize.css' %}
                            {% bundler_embed 'components/Button.tsx' %}
                            {% bundler_embed 'login.css.ts' %}
                        """
                        ).render(context)
                        tpl = Template("{% load bundler %}{% bundler_embed_collected_assets %}")
                        actual = asset_context.post_process(tpl.render(context))
                        self.assertEqual(
                            expected,
                            actual,
                        )

    def test_bundler_embed_collected_assets_content_type_css(self):
        for bundler_name, expected in [
            (
                "test_development_bundler",
                # In dev we don't know about any dependencies as everything is loaded by the JS file. As such requesting
                # css will result in no embeds
                "",
            ),
            ("test_production_bundler", '<link rel="stylesheet" href="/static/assets/Button-abc123.css">'),
        ]:
            with self.subTest(bundler_name=bundler_name):
                with BundlerAssetContext(
                    frontend_asset_registry=bypass_frontend_asset_registry
                ) as asset_context:
                    with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                        context = Context()
                        Template(
                            "{% load bundler %}{% bundler_embed 'components/Button.tsx' content_type='text/css' %}"
                        ).render(context)
                        tpl = Template("{% load bundler %}{% bundler_embed_collected_assets %}")
                        actual = asset_context.post_process(tpl.render(context))
                        self.assertEqual(
                            expected,
                            actual,
                        )

    def test_bundler_embed_collected_assets_content_type_js(self):
        for bundler_name, expected in [
            (
                "test_development_bundler",
                f'<script src="{self.dev_url}components/Button.tsx" type="module"></script>',
            ),
            (
                "test_production_bundler",
                '<script src="/static/assets/Button-def456.js" type="module"></script>',
            ),
        ]:
            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry) as asset_context:
                with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                    context = Context()
                    Template(
                        "{% load bundler %}{% bundler_embed 'components/Button.tsx' content_type='text/javascript' %}"
                    ).render(context)
                    tpl = Template("{% load bundler %}{% bundler_embed_collected_assets %}")
                    actual = asset_context.post_process(tpl.render(context))
                    self.assertEqual(
                        expected,
                        actual,
                    )

    def test_bundler_embed_collected_assets_css_inline(self):
        for bundler_name, expected in [
            (
                "test_development_bundler",
                # Dev currently doesn't support inline css so will just output script tag
                f'<script src="{self.dev_url}components/Button.tsx" type="module"></script>',
            ),
            (
                "test_production_bundler",
                '<script src="/static/assets/Button-def456.js" type="module"></script>\n'
                f'<style>{inline_css_prod[fixtures_dir / "build_test/assets/Button-abc123.css"]}</style>',
            ),
        ]:
            with self.subTest(bundler_name=bundler_name):
                with BundlerAssetContext(
                    frontend_asset_registry=bypass_frontend_asset_registry,
                    html_target=HtmlGenerationTarget("test", include_scripts=True, inline_css=True),
                ) as asset_context:
                    with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                        context = Context()
                        Template("{% load bundler %}{% bundler_embed 'components/Button.tsx' %}").render(
                            context
                        )
                        tpl = Template("{% load bundler %}{% bundler_embed_collected_assets %}")
                        with mock.patch("pathlib.Path.read_text", mock_read_text):
                            actual = asset_context.post_process(tpl.render(context))
                            self.assertEqual(
                                expected,
                                actual,
                            )

    def test_bundler_embed_collected_assets_no_scripts(self):
        for bundler_name, expected in [
            (
                "test_development_bundler",
                "",
            ),
            ("test_production_bundler", '<link rel="stylesheet" href="/static/assets/Button-abc123.css">'),
        ]:
            with BundlerAssetContext(
                frontend_asset_registry=bypass_frontend_asset_registry,
                html_target=HtmlGenerationTarget("test", include_scripts=False, inline_css=False),
            ) as asset_context:
                with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                    context = Context()
                    Template("{% load bundler %}{% bundler_embed 'components/Button.tsx' %}").render(context)
                    tpl = Template("{% load bundler %}{% bundler_embed_collected_assets %}")
                    actual = asset_context.post_process(tpl.render(context))
                    self.assertEqual(
                        expected,
                        actual,
                    )


# Set these so resolving css-mappings directories
@override_ap_frontend_settings(CACHE_DIR=fixtures_dir, PRODUCTION_DIR=fixtures_dir)
@mock.patch(
    "alliance_platform.frontend.bundler.vanilla_extract.VanillaExtractClassMapping._create_mapping",
    mock_create_mapping,
)
class TestVanillaExtractTemplateTag(SimpleTestCase):
    def setUp(self) -> None:
        self.test_production_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="production",
        )
        self.test_development_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="development",
        )
        self.dev_url = self.test_development_bundler.dev_server_url

    def test_stylesheet(self):
        for bundler_name, expected in [
            # Resolved from fixtures dir / development-css-mappings
            ("test_development_bundler", "LoginView__abc123"),
            # Resolved from fixtures dir / production-css-mappings
            ("test_production_bundler", "__abc123"),
        ]:
            with BundlerAssetContext(
                frontend_asset_registry=bypass_frontend_asset_registry, skip_checks=True
            ):
                with override_ap_frontend_settings(BUNDLER=getattr(self, bundler_name)):
                    tpl = Template(
                        "{% load vanilla_extract %}"
                        "{% stylesheet 'login.css.ts' as styles %}\n"
                        "{{ styles.LoginView }}"
                    )
                    context = Context()
                    actual = tpl.render(context).split("\n")[-1].strip()
                    self.assertEqual(
                        expected,
                        actual,
                    )


@override_ap_frontend_settings(DEBUG_COMPONENT_OUTPUT=False)
class TestComponentTemplateTagCodeGen(SimpleTestCase):
    def setUp(self) -> None:
        self.test_production_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="production",
        )

        self.test_development_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="development",
        )
        self.dev_url = self.test_development_bundler.dev_server_url

    # TODO: Very open to ideas on how to better test this. This will be rather fragile to any
    # changes to how the JS code is generated.
    def assertCodeEqual(self, expected, actual):
        expected = format_code(expected)
        actual = format_code(actual)
        return self.assertEqual(
            [line.strip() for line in expected.splitlines() if line.strip()],
            [line.strip() for line in actual.splitlines() if line.strip()],
        )

    @override_settings(STATIC_URL="/static/")
    def test_component_simple_render(self):
        for bundler_name, expected in [
            # Resolved from fixtures dir / development-css-mappings
            (
                "test_development_bundler",
                f"""
                    <dj-component data-djid="C1"><!-- ___SSR_PLACEHOLDER_0___ --></dj-component>
                    <script type="module">
                        import {{ createElement, renderComponent }} from '{self.dev_url}frontend/src/renderComponent.tsx';
                        import Button, {{  }} from '{self.dev_url}components/Button.tsx';
                        renderComponent(
                          document.querySelector("[data-djid='C1']"),
                          createElement(Button, {{}}, "Click Me"),
                          "C1",
                          true
                        );
                    </script>
                """,
            ),
            # Resolved from fixtures dir / production-css-mappings
            (
                "test_production_bundler",
                """
                    <dj-component data-djid="C1"><!-- ___SSR_PLACEHOLDER_0___ --></dj-component>
                    <script type="module">
                        import { createElement, renderComponent } from '/static/assets/renderComponent-e1.js';
                        import Button, {  } from '/static/assets/Button-def456.js';
                        renderComponent(
                            document.querySelector("[data-djid='C1']"),
                            createElement(Button, {}, "Click Me"),
                            "C1",
                            true
                        );
                    </script>
                """,
            ),
        ]:
            with mock.patch(
                "alliance_platform.frontend.bundler.middleware.BundlerAssetContext.generate_id"
            ) as mock_method:
                container_id = "C1"
                mock_method.return_value = container_id
                with BundlerAssetContext(
                    frontend_asset_registry=bypass_frontend_asset_registry,
                    skip_checks=True,
                ) as asset_context:
                    with override_ap_frontend_settings(
                        BUNDLER=getattr(self, bundler_name),
                    ):
                        tpl = Template(
                            "{% load react %}"
                            "{% component 'components/Button.tsx' %}Click Me{% endcomponent %}"
                        )
                        context = Context()
                        actual = tpl.render(context)
                        self.assertCodeEqual(expected, actual)
                        self.assertEqual(len(asset_context.ssr_queue), 1)

    def test_component_children(self):
        with override_ap_frontend_settings(
            BUNDLER=self.test_development_bundler,
        ):
            with mock.patch(
                "alliance_platform.frontend.bundler.middleware.BundlerAssetContext.generate_id"
            ) as mock_method:
                container_id = "C1"
                mock_method.return_value = container_id
                with BundlerAssetContext(
                    skip_checks=True, frontend_asset_registry=bypass_frontend_asset_registry
                ) as asset_context:
                    tpl = Template(
                        "{% load react %}"
                        "{% component 'components/Button.tsx' %}Click {% component 'strong' %}Me{% endcomponent %}{% endcomponent %}"
                    )
                    context = Context()
                    actual = tpl.render(context)
                    self.assertCodeEqual(
                        actual,
                        """
                        <dj-component data-djid="C1"><!-- ___SSR_PLACEHOLDER_0___ --></dj-component>
                        <script type="module">
                            import { createElement, renderComponent } from "%sfrontend/src/renderComponent.tsx";
                            import Button from "%scomponents/Button.tsx";

                            renderComponent(
                              document.querySelector("[data-djid='C1']"),
                              createElement(Button, {}, "Click ", createElement("strong", {}, "Me")),
                              "C1",
                              true
                            );
                        </script>
                        """
                        % (
                            self.dev_url,
                            self.dev_url,
                        ),  # used this rather than f-string to avoid escaping curly braces
                    )
                    ssr_context = SSRSerializerContext(self.test_development_bundler)
                    items = json.loads(
                        json.dumps(
                            {
                                placeholder: ssr_item.serialize(ssr_context)
                                for placeholder, ssr_item in asset_context.ssr_queue.items()
                            },
                            ssr_context=ssr_context,
                            cls=SSRJsonEncoder,
                        )
                    )
                    self.assertEqual(len(ssr_context.get_required_imports()), 1)
                    self.assertEqual(len(asset_context.ssr_queue), 1)
                    import_id = next(iter(ssr_context.get_required_imports().keys()))
                    self.assertEqual(
                        items,
                        {
                            "<!-- ___SSR_PLACEHOLDER_0___ -->": {
                                "ssrType": "Component",
                                "payload": {
                                    "component": [
                                        "@@CUSTOM",
                                        "ComponentImport",
                                        {
                                            "import": import_id,
                                            "propertyName": None,
                                        },
                                    ],
                                    "props": {
                                        "children": [
                                            "Click ",
                                            [
                                                "@@CUSTOM",
                                                "Component",
                                                {"component": "strong", "props": {"children": "Me"}},
                                            ],
                                        ]
                                    },
                                    "identifierPrefix": "C1",
                                },
                            }
                        },
                    )

    def test_ssr_disabled(self):
        with override_ap_frontend_settings(
            BUNDLER=self.test_development_bundler,
        ):
            with BundlerAssetContext(
                frontend_asset_registry=bypass_frontend_asset_registry, skip_checks=True
            ) as asset_context:
                tpl = Template(
                    "{% load react %}"
                    "{% component 'components/Button.tsx' ssr:disabled=True %}Click {% component 'strong' %}Me{% endcomponent %}{% endcomponent %}"
                )
                context = Context()
                tpl.render(context)
                self.assertFalse(asset_context.ssr_queue)

    def test_props_dict(self):
        """Tests that props passed through as a dict to `props` kwarg work"""
        with override_ap_frontend_settings(
            BUNDLER=self.test_development_bundler,
        ):
            with BundlerAssetContext(
                frontend_asset_registry=bypass_frontend_asset_registry, skip_checks=True
            ):
                with mock.patch(
                    "alliance_platform.frontend.bundler.middleware.BundlerAssetContext.generate_id"
                ) as mock_method:
                    container_id = "C1"
                    mock_method.return_value = container_id
                    tpl = Template(
                        "{% load react %}"
                        "{% component 'button' disabled=True props=props %}Click{% endcomponent %}"
                    )
                    context = Context(
                        {
                            "props": {
                                "aria-label": "Click Me",
                                "date": datetime.date(2022, 12, 1),
                            }
                        }
                    )
                    actual = tpl.render(context)
                    self.assertIn(
                        """renderComponent(document.querySelector("[data-djid='C1']"), createElement("button", {disabled: true, "aria-label": "Click Me", date: new CalendarDate(2022, 12, 1)}, "Click"), "C1", true)""",
                        actual,
                    )

    def assertSerializedPropsEqual(self, template_contents: str, context: dict, expected_props: dict):
        """Helper to compare props generated for a components

        Args:
            template_contents: Template string to use. Should include 1 {% component %} tag that will have it's
                props extracted
            context: The context to pass to the template. It will be wrapped in ``Context`` for you.
            expected_props: The expected props for the component
        """
        node: ComponentNode = cast(
            ComponentNode, Template(template_contents).compile_nodelist().get_nodes_by_type(ComponentNode)[0]
        )
        props = node.resolve_props(Context(context))
        ssr_context = SSRSerializerContext(self.test_development_bundler)
        props = json.loads(
            json.dumps(props.serialize(ssr_context), cls=SSRJsonEncoder, ssr_context=ssr_context)
        )

        self.assertEqual(expected_props, props)

    def test_deep_nesting(self):
        """Tests that props passed through as a dict to `props` kwarg work"""
        with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry, skip_checks=True):
            self.assertSerializedPropsEqual(
                "{% load react %}"
                "{% component 'button' disabled=True props=props %}Click "
                "{% if True %}{% component 'div' %}Me "
                "{% if True %}{% component 'span' %}Now {% endcomponent %}{% endif %}{% endcomponent %}{% endif %}"
                " Please, "
                "{% component 'strong' %}Thanks!{% endcomponent %}"
                "{% endcomponent %}",
                {
                    "props": {
                        "aria-label": "Click Me",
                    }
                },
                {
                    "disabled": True,
                    "children": [
                        "Click ",
                        [
                            "@@CUSTOM",
                            "Component",
                            {
                                "component": "div",
                                "props": {
                                    "children": [
                                        "Me ",
                                        [
                                            "@@CUSTOM",
                                            "Component",
                                            {
                                                "component": "span",
                                                "props": {"children": "Now "},
                                            },
                                        ],
                                    ],
                                },
                            },
                        ],
                        " Please, ",
                        [
                            "@@CUSTOM",
                            "Component",
                            {
                                "component": "strong",
                                "props": {"children": "Thanks!"},
                            },
                        ],
                    ],
                    "aria-label": "Click Me",
                },
            )

    def test_date_prop(self):
        with override_ap_frontend_settings(
            BUNDLER=self.test_development_bundler,
        ):
            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry):
                self.assertSerializedPropsEqual(
                    "{% load react %}" "{% component 'DatePicker' date=date %}{% endcomponent %}",
                    {
                        "date": datetime.date(2022, 12, 1),
                    },
                    {"date": ["@@CUSTOM", "Date", [2022, 12, 1]]},
                )

                self.assertSerializedPropsEqual(
                    "{% load react %}" "{% component 'DateTimePicker' datetime=datetime %}{% endcomponent %}",
                    {
                        "datetime": datetime.datetime(2022, 12, 1, 12, 10, 10, 0),
                    },
                    {"datetime": ["@@CUSTOM", "DateTime", [2022, 12, 1, 12, 10, 10, 0]]},
                )

                self.assertSerializedPropsEqual(
                    "{% load react %}" "{% component 'DateTimePicker' datetime=datetime %}{% endcomponent %}",
                    {
                        "datetime": make_aware(datetime.datetime(2022, 12, 1, 12, 10, 10, 0)),
                    },
                    {
                        "datetime": [
                            "@@CUSTOM",
                            "ZonedDateTime",
                            [2022, 12, 1, "Australia/Melbourne", 39600000, 12, 10, 10, 0],
                        ]
                    },
                )

    def test_time_prop(self):
        with override_ap_frontend_settings(
            BUNDLER=self.test_development_bundler,
        ):
            with BundlerAssetContext(frontend_asset_registry=bypass_frontend_asset_registry):
                self.assertSerializedPropsEqual(
                    "{% load react %}" "{% component 'Time' time=time %}{% endcomponent %}",
                    {
                        "time": datetime.time(12, 30, 15, 5),
                    },
                    {
                        "time": [
                            "@@CUSTOM",
                            "Time",
                            [12, 30, 15, 5],
                        ]
                    },
                )

    def test_whitespace_handling(self):
        with override_ap_frontend_settings(BUNDLER=self.test_development_bundler):
            with mock.patch(
                "alliance_platform.frontend.bundler.middleware.BundlerAssetContext.generate_id"
            ) as mock_method:
                container_id = "C1"
                mock_method.return_value = container_id
                with BundlerAssetContext(
                    skip_checks=True, frontend_asset_registry=bypass_frontend_asset_registry
                ):
                    tpl_str = """
                        {% load react %}
                        {% component 'components/Button.tsx' %}

                            Click {% component 'strong' %} Me  {% endcomponent %}
                            Ok
                                then
Some               space
   {% component 'b' %}Test
         {{ ok }} {{ num }}   {% endcomponent %}
                        {% endcomponent %}"
                    """
                    self.assertSerializedPropsEqual(
                        tpl_str,
                        {"ok": True, "num": 5},
                        {
                            "children": [
                                "Click ",
                                [
                                    "@@CUSTOM",
                                    "Component",
                                    {"component": "strong", "props": {"children": " Me  "}},
                                ],
                                "Ok then Some               space",
                                [
                                    "@@CUSTOM",
                                    "Component",
                                    {"component": "b", "props": {"children": "Test True 5   "}},
                                ],
                            ]
                        },
                    )


test_development_bundler = TestViteBundler(
    **bundler_kwargs,  # type: ignore[arg-type]
    mode="development",
)


@override_ap_frontend_settings(
    # We rely on this in _get_debug_tree
    DEBUG_COMPONENT_OUTPUT=True,
    BUNDLER=test_development_bundler,
)
class TestComponentTemplateTagOutput(SimpleTestCase):
    """For ease of testing, these tests just use the pretty debug output for comparing code

    For testing the actual codegen itself, see the above tests
    """

    def setUp(self):
        self.bundler_context = BundlerAssetContext(
            frontend_asset_registry=bypass_frontend_asset_registry, skip_checks=True
        )
        self.bundler_context.__enter__()

    def tearDown(self):
        self.bundler_context.__exit__(None, None, None)

    def _get_debug_tree(self, template_contents: str, **kwargs: dict):
        def patch_debug_tree(self, props: ComponentProps, include_template_origin=True):
            # Like the default implementation except we never include template origin string
            printer = TypescriptPrinter(jsx_transform=None)
            generator = ComponentSourceCodeGenerator(self)
            jsx_element = generator.create_jsx_element_node(self, props, False)
            return printer.print(jsx_element)

        with mock.patch(
            "alliance_platform.frontend.templatetags.react.ComponentNode.print_debug_tree", patch_debug_tree
        ):
            template = Template("{% load react %}" + template_contents)
            context = Context(kwargs)
            context.template = template
            contents = template.render(context)
            code = contents.split("<!--").pop().split("-->").pop(0)
            return code.strip()

    def assertComponentEqual(self, template_code, expected_output, **kwargs):
        self.assertEqual(
            run_prettier(self._get_debug_tree(template_code, **kwargs)),
            run_prettier(expected_output),
        )

    def test_django_lazy_object_as_prop(self):
        def get_token():
            return "abc123"

        token = lazy(get_token, str)
        self.assertComponentEqual(
            """{% component "CustomForm" csrf_token=csrf_token %}{% endcomponent %}""",
            """
            <CustomForm csrfToken="abc123" />
            """,
            csrf_token=token,
        )

    def test_django_simple_lazy_object_as_prop(self):
        def get_token():
            return "abc123"

        token = SimpleLazyObject(get_token)
        self.assertComponentEqual(
            """{% component "CustomForm" csrf_token=csrf_token %}{% endcomponent %}""",
            """
            <CustomForm csrfToken="abc123" />
            """,
            csrf_token=token,
        )

    def test_variable_string_concat(self):
        self.assertComponentEqual(
            """{% component "Button" %}Hello {{ name }}{% endcomponent %}""",
            """
            <Button>
              Hello Sam
            </Button>
            """,
            name="Sam",
        )

    def test_boolean_attribute(self):
        self.assertComponentEqual(
            """{% component "Button" is_disabled=True %}Disabled{% endcomponent %}""",
            """<Button isDisabled={true}>Disabled</Button>""",
        )

    def test_string_attribute(self):
        self.assertComponentEqual(
            """{% component "Button" variant="primary" %}Click Me{% endcomponent %}""",
            """<Button variant="primary">Click Me</Button>""",
        )

    def test_numeric_attribute(self):
        self.assertComponentEqual(
            """{% component "Range" min_value=5  %}{% endcomponent %}""",
            """<Range minValue={5} />""",
        )

    def test_variable_attribute(self):
        self.assertComponentEqual(
            """{% component "Button" variant=variant %}Click Me{% endcomponent %}""",
            """<Button variant="secondary">Click Me</Button>""",
            variant="secondary",
        )

    def test_variable_attribute_raw_html(self):
        self.assertComponentEqual(
            """{% component "Button" %}Click <strong id="test-{{id}}">Me</strong>{% endcomponent %}""",
            """<Button>Click <strong id="test-secondary">Me</strong></Button>""",
            id="secondary",
        )

    def test_special_attribute(self):
        self.assertComponentEqual(
            """{% component "Button" aria-label="Disable" %}X{% endcomponent %}""",
            """<Button aria-label="Disable">X</Button>""",
        )

    def test_html_entities(self):
        self.assertComponentEqual(
            """{% component "Button" %}&ldquo;Testing and Stuff&rdquo;{% endcomponent %}""",
            """<Button>“Testing and Stuff”</Button>""",
        )

    def test_nested_component(self):
        self.assertComponentEqual(
            """{% component "Button" %}Welcome {% component "strong" %}{{ name }}{% endcomponent %}{% endcomponent %}""",
            """<Button>Welcome <strong>Sam</strong></Button>""",
            name="Sam",
        )

    def test_nested_component_raw_html(self):
        self.assertComponentEqual(
            """{% component "Button" %}Welcome <strong>{{ name }}</strong>{% endcomponent %}""",
            """
            <Button>Welcome <strong>Sam</strong></Button>
            """,
            name="Sam",
        )

    def test_nested_component_raw_html_from_include(self):
        self.assertComponentEqual(
            """{% component "Button" %}Welcome {% include "react_test_templates/nested_raw_tag.html" with name=name %}{% endcomponent %}""",
            """
            <Button>Welcome <strong>Sam</strong></Button>
            """,
            name="Sam",
        )

    def test_nested_component_from_include(self):
        self.assertComponentEqual(
            """{% component "Button" %}Welcome {% include "react_test_templates/nested_tag.html" with name=name %}{% endcomponent %}""",
            """
            <Button>Welcome <strong>Sam</strong></Button>
            """,
            name="Sam",
        )

    def test_template_include(self):
        self.assertComponentEqual(
            """{% component "Button" %}Hello {% include "react_test_templates/simple_include.html" %}{% endcomponent %}""",
            """
            <Button>
              Hello Inner Content
            </Button>
            """,
            name="Sam",
        )

    def test_pass_component_as_prop(self):
        self.assertComponentEqual(
            """
            {% component "Icon" as icon %}{% endcomponent %}
            {% component "Button" icon=icon %}Click{% endcomponent %}""",
            """<Button icon={<Icon />}>Click</Button>""",
            name="Sam",
        )

    def test_html_in_var(self):
        self.assertComponentEqual(
            """
            {% component "div" %}{{ text }}{% endcomponent %}""",
            """<div><strong>Should not be escaped</strong></div>""",
            text=mark_safe("<strong>Should not be escaped</strong>"),
        )

        self.assertComponentEqual(
            """
            {% component "div" %}{{ text }}{% endcomponent %}""",
            """<div>{'<strong>Should be escaped</strong>'}</div>""",
            text="<strong>Should be escaped</strong>",
        )



from __future__ import annotations

from dataclasses import dataclass
import json
from json import JSONDecodeError
import logging
from pathlib import Path
import uuid

from allianceutils.util import camelize
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
import requests

from . import get_bundler
from .base import BaseBundler

logger = logging.getLogger("alliance_platform.frontend")

SSR_FAILURE_PLACEHOLDER = "<!-- SSR_FAILED -->"


class SSRJsonEncoder(DjangoJSONEncoder):
    """Custom encoder that handles ``SSRSerializable`` objects

    Any ``SSRSerializable`` object will be serialized using the ``serialize`` method, and will
    have access to the ``SSRSerializerContext`` object.
    """

    ssr_context: SSRSerializerContext

    def __init__(self, ssr_context: SSRSerializerContext, **kwargs):
        self.ssr_context = ssr_context
        super().__init__(**kwargs)

    def default(self, o):
        if isinstance(o, SSRSerializable):
            return o.serialize(self.ssr_context)
        return super().default(o)


class SSRSerializerContext:
    """Context available to ``SSRSerializable`` objects during serialization"""

    required_imports: dict[ImportDefinition, tuple[str, dict]]
    bundler: BaseBundler

    def __init__(self, bundler: BaseBundler):
        self.required_imports = {}
        self.prefix = uuid.uuid4().hex
        self.bundler = bundler

    def add_import(self, definition: ImportDefinition) -> str:
        """Add the specified import to the list of required imports and return the cache key

        This allows us to collect all the imports that are required for a given SSR request
        and load them all up front, then assign them as part of the JSON parsing.

        For example, an import will be serialized as:

            ['@@CUSTOM', 'ModuleImport', 'abc123']

        ``abc123`` is the key returned by ``add_import``. In ``processSSRRequest``, the required
        imports will be loaded first, then the JSON will be parsed with a custom reviver. This
        reviver will replace any ``['@@CUSTOM', 'ModuleImport', 'abc123']`` with the actual resolved
        import.

        This turns out to be a lot faster than traversing the object after JSON parsing and loading any
        encountered imports.
        """
        if definition not in self.required_imports:
            cache_key = f"{self.prefix}__{len(self.required_imports)}"
            path = definition.path
            self.required_imports[definition] = (
                cache_key,
                {
                    "path": str(
                        self.bundler.resolve_ssr_import_path(
                            self.bundler.validate_path(path, resolve_extensions=[".ts", ".tsx"])
                        )
                    ),
                    "import_name": definition.import_name,
                    "is_default_import": definition.is_default_import,
                },
            )
        return self.required_imports[definition][0]

    def get_required_imports(self) -> dict[str, dict]:
        return {key: camelize(definition) for key, definition in self.required_imports.values()}


class SSRSerializable:
    """Mixin for classes which can be serialized for SSR

    ``SSRJsonEncoder`` will handle converting the object to JSON in the form::

        ["@@CUSTOM", <tag name>, <serialized representation>]

    For example, a date object might look like:

        ["@@CUSTOM", "date", "2023-01-01"]

    If ``get_tag`` returns `"json"` then the value will be serialized as is and not in the above format. This
    is useful to provide serialization logic for standard JSON serializable objects that require no special
    handling on the frontend.
    """

    def serialize(self, context: SSRSerializerContext) -> dict | str | list:
        """Serialize the object for SSR

        This method should return a value that can be JSON serialized. Note that it's fine to return values
        that contain ``SSRSerializable`` objects, as these will be serialized recursively.
        """
        raise NotImplementedError(f"{self.__class__.__name__} must implement serialize()")


class SSRCustomFormatSerializable(SSRSerializable):
    def get_tag(self):
        """Get the tag to identify the type of serializable. This is matched in nodejs for de-serialization."""
        raise NotImplementedError(f"{self.__class__.__name__} must implement get_tag()")

    def get_representation(self, context: SSRSerializerContext) -> dict | str | list:
        """Get the representation of the object to be serialized

        This can be any JSON serializable value. The matching reviver in ``ssrJsonReviviers.tsx`` will be passed this value.
        """
        raise NotImplementedError(f"{self.__class__.__name__} must implement get_representation()")

    def serialize(self, context: SSRSerializerContext) -> list:
        tag = self.get_tag()
        return ["@@CUSTOM", tag, self.get_representation(context)]


@dataclass(frozen=True)
class ImportDefinition(SSRCustomFormatSerializable):
    """Describes an import that needs to be resolved.

    This uses ``SSRSerializerContext`` to queue its import. ``BundlerAssetServerSideRenderer.process`` will then
    serialize these imports and pass them to the SSR renderer. ``processSSRRequest`` will then load the imports
    first before de-serializing the JSON for all items in the SSR. In practice, resolving each import once upfront,
    was significantly faster than resolving each import as it was encountered in the JSON.
    """

    path: Path
    import_name: str
    is_default_import: bool

    def get_tag(self):
        return "ModuleImport"

    def get_representation(self, context: SSRSerializerContext) -> str:
        return context.add_import(self)


class SSRItem:
    """
    Base class for server side rendering (SSR) items

    The ``SSRItem`` class is the base class for server side rendering items. An ``SSRItem`` is made up of a type,
    which tells the SSR JavaScript code how to handle the item, and a payload, which serializes any necessary data from
    the server to the frontend.

    For example, a React component payload would encode the details of the component (its name and where to import it from)
    and the props used to render it.

    Each concrete implementation of ``SSRItem`` needs to implement :meth:`~alliance_platform.frontend.bundler.context.SSRItem.get_ssr_type`
    and :meth:`~alliance_platform.frontend.bundler.context.SSRItem.get_ssr_payload`.

    ``SSRItems`` are queued for rendering using the :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.queue_ssr`
    method, and then rendered by :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware`.
    """

    def get_ssr_type(self):
        """Returns the type of SSR which is then interpreted by the frontend

        See ``ssr.ts`` for where this is done
        """
        raise NotImplementedError

    def get_ssr_payload(self, ssr_context: SSRSerializerContext) -> dict:
        """Return the payload required by the frontend for this item

        This should be a JSON serializable Dict
        """
        raise NotImplementedError

    def serialize(self, ssr_context: SSRSerializerContext) -> dict:
        """Serializes to a dict that can be JSON encoded and sent to the frontend

        You shouldn't need to override this method - instead implement ``get_ssr_type`` and
        ``get_ssr_payload``.

        Returns a Dict with two keys: ``ssrType`` and ``payload``.
        """
        return {
            "ssrType": self.get_ssr_type(),
            "payload": self.get_ssr_payload(ssr_context),
        }


class BundlerAssetServerSideRenderer:
    ssr_queue: dict[str, SSRItem]

    def __init__(self, ssr_queue: dict[str, SSRItem]):
        self.ssr_queue = ssr_queue

    def process_ssr(self, payload: dict) -> dict | None:
        """Process an SSR request

        This calls the endpoint specified by :meth:`~alliance_platform.frontend.bundler.base.BaseBundler.get_ssr_url`.

        For the Vite implementation this endpoint is defined in ``dev-server.ts`` for dev, or ``production-ssr-server.ts``
        in preview or production.
        """

        # If dev the rendering happens in dev-server.ts
        bundler = get_bundler()
        ssr_url = bundler.get_ssr_url()
        if not ssr_url:
            logger.error("Can not perform SSR, no SSR URL defined. Set PRODUCTION_SSR_URL env var.")
            return None
        try:
            json_payload = json.dumps(payload, cls=DjangoJSONEncoder)
            ssr_response = requests.post(
                ssr_url,
                data=json_payload,
                headers={"Content-Type": "application/json", **bundler.get_ssr_headers()},
                timeout=1,
            )
            if ssr_response.status_code != 200:
                try:
                    # If SSR server responds with JSON it will have an 'error' and optional 'stack' key for more info
                    data = ssr_response.json()
                    msg = f"Bad response {ssr_response.status_code} from dev-server.ts: {data['error']}"
                    if "stack" in data:
                        msg += f"\n{data['stack']}"
                    logger.error(msg)
                except JSONDecodeError:
                    logger.error(
                        f"Bad response {ssr_response.status_code} from dev-server.ts: {ssr_response.content.decode()}"
                    )
            else:
                # See ssr.ts ServerRenderComponentReturnValue for what this is
                try:
                    return ssr_response.json()
                except JSONDecodeError:
                    logger.error(
                        f"Failed to decode JSON from server rendering, content received: {ssr_response.content.decode()}"
                    )
        except TypeError:
            logger.exception(f"Failed to encode JSON for server rendering. Payload was: {payload}")
        except requests.exceptions.Timeout:
            logger.error("Timed out connecting to vite server for rendering")
        except requests.exceptions.ConnectionError:
            logger.error("Failed to connect to vite server for rendering - is it running?")
        return None

    def process(self, content: str | bytes, global_context: dict):
        if not self.ssr_queue:
            return content
        ssr_context = SSRSerializerContext(get_bundler())
        # items are serialized first so that `ssr_context` can be populated with the required imports
        serialized_items = {
            placeholder: ssr_item.serialize(ssr_context) for placeholder, ssr_item in self.ssr_queue.items()
        }
        items = json.dumps(
            serialized_items,
            ssr_context=ssr_context,
            cls=SSRJsonEncoder,
        )
        # This should match `ServerRenderRequest` in `ssr.ts`
        payload = {
            "itemsJson": items,
            "requiredImports": ssr_context.get_required_imports(),
            "globalContext": global_context,
        }
        # See ssr.ts `renderItems` for more details on what the return value is
        # All requested items will be in either the `renderedItems` Dict or the `errors` Dict.
        # Example return:
        # {
        #    renderedItems: {
        #      "___SSR_PLACEHOLDER_0___": {
        #        "html": "<div>Test</div>",
        #        "renderErrors": ["some error that didn't stop SSR"],
        #      }
        #    },
        #    errors: {},
        #  }
        data = self.process_ssr(payload)
        is_bytes = isinstance(content, bytes)

        def replace_content(placeholder, value):
            if is_bytes:
                return content.replace(placeholder.encode(), value.encode())
            return content.replace(placeholder, value)

        if data:
            for placeholder, value in data["renderedItems"].items():
                content = replace_content(placeholder, value["html"])
                if value.get("renderErrors"):
                    try:
                        debug_component_name = (
                            serialized_items[placeholder]["payload"]["component"].import_name
                            + f' (HTML placeholder="{placeholder}")'
                        )
                    except (KeyError, AttributeError):
                        debug_component_name = f'HTML placeholder="{placeholder}"'
                        pass
                    if settings.DEBUG:
                        # Outputting the errors here is very difficult to read - direct the user to the console instead
                        logger.warning(
                            f"SSR succeeded with errors for item {debug_component_name}. See `yarn dev` console for details."
                        )
                    else:
                        logger.warning(
                            f"SSR succeeded with errors for item {debug_component_name}: {value['renderErrors']}"
                        )
            for placeholder, value in data["errors"].items():
                logger.error(f"Server rendering failed for {self.ssr_queue[placeholder]}: {value}")
                content = replace_content(placeholder, SSR_FAILURE_PLACEHOLDER)
        else:
            for placeholder in serialized_items.keys():
                content = replace_content(placeholder, SSR_FAILURE_PLACEHOLDER)

        return content



from __future__ import annotations

from django import template
from django.template import Library

from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


def table(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Table", False, parser.origin),
    )


def table_header(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "TableHeader", False, parser.origin),
    )


def table_body(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "TableBody", False, parser.origin),
    )


def column_header_link(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source(
            "@alliancesoftware/ui", "ColumnHeaderLink", False, parser.origin
        ),
    )


def row(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Row", False, parser.origin),
    )


def column(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Column", False, parser.origin),
    )


def cell(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Cell", False, parser.origin),
    )


def register_table(register: Library):
    register.tag("Table")(table)
    register.tag("TableHeader")(table_header)
    register.tag("TableBody")(table_body)
    register.tag("ColumnHeaderLink")(column_header_link)
    register.tag("Column")(column)
    register.tag("Row")(row)
    register.tag("Cell")(cell)



from alliance_platform.frontend.settings import ap_frontend_settings
from django.apps.config import AppConfig


class AlliancePlatformFrontendConfig(AppConfig):
    name = "alliance_platform.frontend"
    verbose_name = "Alliance Platform Frontend"
    label = "alliance_platform_frontend"

    def ready(self):
        ap_frontend_settings.check_settings()



import warnings

from django import template
from django.template import Context
from django.template import Library
from django.utils.dateparse import parse_date
from django.utils.dateparse import parse_datetime

from ..templatetags.react import ComponentNode
from ..templatetags.react import ComponentProps
from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


class DatePickerNode(ComponentNode):
    def resolve_props(self, context: Context) -> ComponentProps:
        values = super().resolve_props(context)
        granularity = values.props.get("granularity", "day")
        if granularity not in {"day", "hour", "minute", "second"}:
            warnings.warn(
                f"Invalid granularity '{granularity}' passed to DatePicker. Defaulting to 'day'. If this is a datetime picker this will break."
            )
            granularity = "day"
            values.update({"granularity": self.resolve_prop(granularity, context)})
        if values.has_prop("default_value") and isinstance(values.props["default_value"], str):
            values.update(
                {
                    "default_value": self.resolve_prop(
                        (parse_date if granularity == "day" else parse_datetime)(
                            values.props["default_value"]
                        ),
                        context,
                    )
                }
            )
        return values


def date_picker(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "DatePicker", False, parser.origin),
        node_class=DatePickerNode,
    )


def register_date_picker(register: Library):
    register.tag("DatePicker")(date_picker)



from django import template
from django.template import Library

from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


def button(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Button", False, parser.origin),
    )


def button_group(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "ButtonGroup", False, parser.origin),
    )


def register_button(register: Library):
    register.tag("Button")(button)
    register.tag("ButtonGroup")(button_group)



import logging
from pathlib import Path
import re
import subprocess
from typing import Any

from packaging.specifiers import InvalidSpecifier
from packaging.specifiers import SpecifierSet
from packaging.version import InvalidVersion
from packaging.version import Version

from .dom_possible_standard_names import possible_standard_names

logger = logging.getLogger("alliance_platform.frontend")


def get_node_ver(node_path: str) -> Version | None:
    """
    Get the node verson from a given path

    if raise_errors is not True then exceptions will be swallowed and None will be returned on error
    """
    try:
        proc = subprocess.Popen(
            [node_path, "--version"],
            stdout=subprocess.PIPE,
        )
        stdout, _ = proc.communicate()
        stdout_str = stdout.decode("utf-8").strip()
        returncode = proc.returncode
        if returncode != 0:
            raise ChildProcessError("Non-zero exit code")
        return Version(stdout_str.lstrip("v"))
    except (OSError, InvalidVersion, ChildProcessError):
        return None


def guess_node_path(nvmrc_path: Path) -> str | None:
    # nvm's version strings aren't quite the same as those
    # used by python but the overlap makes it good enough

    nvm_ver = (nvmrc_path).read_text().strip().lstrip("v")
    nvm_ver_re = re.compile(r"^v?" + re.escape(nvm_ver) + r"\b")
    try:
        nvm_specifier = SpecifierSet("==" + nvm_ver + ".*")
    except InvalidSpecifier as invalid_spec:
        raise ValueError(f"Cannot understand .nvmrc version specifier {nvm_ver}") from invalid_spec

    default_ver = get_node_ver("node")
    if default_ver is not None and default_ver in nvm_specifier:
        return "node"

    # default didn't work; try other options
    candidate_paths: list[str | Path] = [
        "/usr/local/bin/node",
        "/usr/bin/node",
    ]

    # add nvm node versions
    paths = Path.home() / ".nvm/versions/node"
    try:
        for p in paths.iterdir():
            if re.match(nvm_ver_re, p.name):
                candidate_paths.append(p / "bin/node")
    except IOError as ioe:
        logger.info(f"Error scanning nvm versions: {ioe}")

    # get version for each node candidate
    candidate_path_vers = {p: get_node_ver(str(p)) for p in candidate_paths}

    # look for the highest version that still matches
    matched_path_vers = [
        (ver, p) for p, ver in candidate_path_vers.items() if ver is not None and ver in nvm_specifier
    ]
    matched_path_vers.sort(reverse=True)

    try:
        return str(matched_path_vers[0][1])
    except IndexError:
        logger.error(f"Can't find node version {nvm_ver}")
        return None

    # TODO: if we can't find an exact match then do we want to fall back to
    # the closest version? (eg can't find v12 so use v13 instead)
    # if not re.match(, default_ver):
    #     warning(f"Wrong node version (expected {expected_ver}, got {default_ver}")


def transform_attribute_names(attrs: dict[str, Any]) -> dict[str, Any]:
    """Transform HTML attributes to use proper cased names for passing to JSX

    Example::

        >>> transform_attribute_names({ "class": "item" })

        { "className": "item" }
    """
    return {possible_standard_names.get(k, k): v for k, v in attrs.items()}



from pathlib import Path
import warnings

from allianceutils.template import is_static_expression
from allianceutils.template import parse_tag_arguments
from django import template
from django.template import Context
from django.template import Origin
from django.template.base import UNKNOWN_SOURCE

from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..bundler.context import BundlerAsset
from ..bundler.vanilla_extract import resolve_vanilla_extract_class_mapping

register = template.Library()


class VanillaExtractStylesheetNode(template.Node, BundlerAsset):
    def __init__(self, filename: Path, origin: Origin | None, attrs=None, css_modules_target_var=None):
        self.filename = filename
        self.attrs = attrs
        self.css_modules_target_var = css_modules_target_var
        super().__init__(origin or Origin(UNKNOWN_SOURCE))

    def get_paths_for_bundling(self) -> list[Path]:
        return [self.filename]

    def render(self, context: Context):
        if self.css_modules_target_var:
            context[self.css_modules_target_var] = resolve_vanilla_extract_class_mapping(
                self.bundler, self.filename
            )
        # TODO: Currently HMR works but it forces a full refresh of the page always. Not clear why.
        # One solution is to write out a temporary `.ts` file that loads the css file. This sort of worked
        # but was a bit flakey so I've removed it for the time being. Needs further investigation into
        # cause.
        items = self.bundler.get_embed_items(self.get_paths_for_bundling())
        for item in items:
            self.bundler_asset_context.queue_embed_file(item)
        # Nothing to render - tags are added to head
        return ""


@register.tag("stylesheet")
def stylesheet(parser: template.base.Parser, token: template.base.Token):
    """
    Add a vanilla extract CSS file the page, optionally exposing class name mapping in a template variable.

    If the CSS file includes exported class names, you can access the mapping by specifying a variable with the syntax
    ``as <var name>``. This functionality relies on the plugin defined by in ``vanillaExtractWithExtras.ts``.

    If you do not specify a variable using the ``as <var name>`` syntax, the styles will only be available globally,
    and any specified variables will be ignored.

    For more information on how paths are resolved, refer to the documentation on :ref:`resolving_paths`.

    The CSS file is not embedded inline where the tag is used, rather it is added by the :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_embed`
    tag.

    Usage::

        {% load vanilla_extract %}

        {% stylesheet "./myView.css.ts" as styles %}

        <div class="{{ styles.section }}">
            <h1 class="{{ styles.heading }}">My View</h1>
            ...
        </div>


    .. note:: If you need to include a plain CSS file use the :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_embed`
            tag instead.

    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)

    if len(args) != 1:
        raise template.TemplateSyntaxError(
            f"'{tag_name}' must receive a single argument which is the name of the stylesheet to include"
        )

    if not is_static_expression(args[0]):
        raise template.TemplateSyntaxError(
            f"'{tag_name}' must be passed a static string for it's argument (encountered variable '{args[0].var}')"
        )
    f = args[0].resolve(Context())
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
    filename = bundler.resolve_path(
        f,
        resolver_context,
        suffix_whitelist=[".css.ts"],
        suffix_hint=f'This tag is for vanilla-extract .css.ts files only. To embed a regular stylesheet just use {{% bundler_embed "{f}" %}}',
    )
    return VanillaExtractStylesheetNode(filename, parser.origin, kwargs, css_modules_target_var=target_var)


@register.filter
def style_key(styles, key):
    if key not in styles:
        warnings.warn(f"Could not find style key '{key}'. Known keys: {', '.join(styles)}")
        return ""
    return styles[key]



from pathlib import Path


class FrontendAssetRegistry:
    """Stores any extra assets in addition to those found in templates that should be included in the frontend build

    This should be populated on startup such that the assets are available when
    :class:`extract_frontend_assets <alliance_platform.frontend.management.commands.extract_frontend_assets.Command>` runs.

    The registry to use is specified by the  :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.FRONTEND_ASSET_REGISTRY`
    setting

    .. warning::

        In most cases you do not need this.

        This is necessary only for assets that can't be auto-discovered - for example in a dynamic template. If you try
        and use an asset that cannot be auto-discovered an error will be raised.

    Usage::

        frontend_asset_registry = FrontendAssetRegistry()

        frontend_asset_registry.add_asset(
            settings.PROJECT_DIR / "frontend/src/file1.tsx",
            settings.PROJECT_DIR / "frontend/src/file2.tsx",
        )

    """

    _assets: set[Path]

    def __init__(self):
        self._assets = set()
        self._locked = False

    def add_asset(self, *filenames: Path):
        """Add asset to be included in frontend build

        Should be a path to file to include
        """
        if self._locked:
            raise ValueError(
                "Cannot add assets to registry after it's locked. Make sure all assets are added at startup."
            )
        for filename in filenames:
            if not filename.is_absolute():
                raise ValueError(
                    f'Filenames should be absolute - e.g. Try `settings.PROJECT_DIR / "{filename}"`'
                )
            if not filename.exists():
                raise ValueError(f"{filename} was added to frontend asset registry but does not exist")
            self._assets.add(filename)

    def get_asset_paths(self) -> set[Path]:
        """Get all assets to include in build"""
        return set(self._assets)

    def get_unknown(self, *filenames: Path) -> list[Path]:
        """From the specified filename(s) return any that aren't in the registry"""
        missing = []
        for filename in filenames:
            if filename not in self._assets:
                missing.append(filename)
        return missing

    def lock(self):
        self._locked = True



from alliance_platform.frontend.settings import AlliancePlatformFrontendSettingsType
from django.conf import settings
from django.test import override_settings
from typing_extensions import Unpack


class override_ap_frontend_settings(override_settings):
    def __init__(self, CACHE_DIR=None, **kwargs: Unpack[AlliancePlatformFrontendSettingsType]):
        CORE = settings.ALLIANCE_PLATFORM.get("CORE", {})
        if CACHE_DIR:
            CORE["CACHE_DIR"] = CACHE_DIR
        super().__init__(
            ALLIANCE_PLATFORM={
                **settings.ALLIANCE_PLATFORM,
                "CORE": CORE,
                "FRONTEND": {**settings.ALLIANCE_PLATFORM.get("FRONTEND", {}), **kwargs},
            }
        )



from django import template
from django.template import Context
from django.template import Library

from ..templatetags.react import ComponentNode
from ..templatetags.react import ComponentProps
from ..templatetags.react import NestedComponentProp
from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


class InlineAlertNode(ComponentNode):
    def resolve_props(self, context: Context) -> ComponentProps:
        props = super().resolve_props(context)
        # If only a string is passed as child then wrap it in a <Content> component
        children = props.props.get("children")
        if isinstance(children, str):
            props.props["children"] = NestedComponentProp(
                ComponentNode(
                    self.origin,
                    get_module_import_source("@alliancesoftware/ui", "Content", False, self.origin),
                    {"children": children},
                ),
                self,
                context,
            )
        return props


def inline_alert(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        node_class=InlineAlertNode,
        asset_source=get_module_import_source("@alliancesoftware/ui", "InlineAlert", False, parser.origin),
    )


def register_inline_alert(register: Library):
    register.tag("InlineAlert")(inline_alert)



from contextlib import contextmanager
from typing import cast

from alliance_platform.frontend.bundler.context import BundlerAssetContext
from allianceutils.auth.permission import AmbiguousGlobalPermissionWarning
from allianceutils.tests.util import warning_filter
from django.contrib.sessions.backends.base import SessionBase
from django.http import HttpRequest
from django.template import Context
from django.template import Template
from django.test import TestCase
from django.test import override_settings
from django.urls import reverse
from test_alliance_platform_frontend.factory import UserFactory
from test_alliance_platform_frontend.models import User

from .test_utils import override_ap_frontend_settings
from .test_utils.bundler import TestViteBundler
from .test_utils.bundler import bundler_kwargs
from .test_utils.bundler import bypass_frontend_asset_registry


@override_ap_frontend_settings(
    DEBUG_COMPONENT_OUTPUT=False,
)
@override_settings(
    AUTHENTICATION_BACKENDS=(
        # This uses test_utils.rules. This works with reverse_if_probably_allowed as it
        # will correctly infer object level permissions without us needing to setup a custom csv permissions
        # for the test cases
        "rules.permissions.ObjectPermissionBackend",
    ),
)
@override_settings(ROOT_URLCONF="test_alliance_platform_frontend.urls")
@warning_filter("ignore", category=AmbiguousGlobalPermissionWarning)
class UrlFilterPermTemplateTagsTestCase(TestCase):
    PERM = "test_utils.link_is_allowed"
    GLOBAL_PERM_URL = "url_with_perm_global"
    OBJECT_PERM_URL = "url_with_perm_object"
    MULTIPLE_ARGS_URL = "url_with_multiple_args"

    def setUp(self) -> None:
        self.test_production_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="production",
        )

        self.test_development_bundler = TestViteBundler(
            **bundler_kwargs,  # type: ignore[arg-type]
            mode="development",
        )
        self.dev_url = self.test_development_bundler.dev_server_url

    def get_privileged_user(self) -> User:
        user = cast(User, UserFactory(is_superuser=True))
        self.assertTrue(user.has_perm(self.PERM))
        return user

    def get_unprivileged_user(self) -> User:
        user = cast(User, UserFactory(is_superuser=False))
        self.assertFalse(user.has_perm(self.PERM))
        return user

    @contextmanager
    def setup_overrides(self):
        with override_ap_frontend_settings(BUNDLER=self.test_development_bundler):
            with BundlerAssetContext(
                skip_checks=True, frontend_asset_registry=bypass_frontend_asset_registry
            ):
                yield

    def test_global_url_with_perm(self):
        user1 = self.get_privileged_user()
        user2 = self.get_unprivileged_user()

        with self.setup_overrides():
            tpl = Template(
                """
            {% load react %}
            {% load alliance_ui %}
            {% component "a" href=perm|url_with_perm %}{% endcomponent %}
            """
            )
            request = HttpRequest()
            request.user = user2
            request.session = SessionBase()
            context = Context({"request": request, "perm": self.GLOBAL_PERM_URL})
            output = tpl.render(context)
            self.assertEqual(output.strip(), "")
            request.user = user1
            output = tpl.render(context)
            url = reverse(self.GLOBAL_PERM_URL)
            self.assertTrue(f'href: "{url}"' in output)

    def test_object_url_with_perm(self):
        user1 = self.get_privileged_user()
        user2 = self.get_unprivileged_user()

        with self.setup_overrides():
            tpl = Template(
                """
            {% load react %}
            {% load alliance_ui %}
            {% component "a" href=perm|url_with_perm:user.pk|with_perm_obj:user %}{% endcomponent %}
            """
            )

            request = HttpRequest()
            request.user = user2
            request.session = SessionBase()
            context = Context({"request": request, "user": user1, "perm": self.OBJECT_PERM_URL})
            output = tpl.render(context)
            self.assertEqual(output.strip(), "")
            request.user = user1
            output = tpl.render(context)
            url = reverse(self.OBJECT_PERM_URL, args=[user1.pk])
            self.assertTrue(f'href: "{url}"' in output)

    def test_object_url_with_perm_with_kwargs(self):
        user1 = self.get_privileged_user()
        user2 = self.get_unprivileged_user()

        with self.setup_overrides():
            tpl = Template(
                """
            {% load react %}
            {% load alliance_ui %}
            {% component "a" href=perm|url_with_perm|with_kwargs:kwargs %}{% endcomponent %}
            """
            )

            request = HttpRequest()
            request.user = user2
            request.session = SessionBase()
            context = Context(
                {
                    "request": request,
                    "kwargs": {"pk": user1.pk},
                    "perm": self.OBJECT_PERM_URL,
                    "user": user1.pk,
                }
            )
            output = tpl.render(context)
            self.assertEqual(output.strip(), "")
            request.user = user1
            with self.assertNumQueries(1):
                # We expect 1 query to lookup the object
                output = tpl.render(context)
                url = reverse(self.OBJECT_PERM_URL, args=[user1.pk])
                self.assertTrue(f'href: "{url}"' in output)

            with self.assertNumQueries(0):
                # If we pass the object then there should be no query
                tpl = Template(
                    """
                {% load react %}
                {% load alliance_ui %}
                {% component "a" href=perm|url_with_perm|with_kwargs:kwargs|with_perm_obj:user %}{% endcomponent %}
                """
                )
                output = tpl.render(context)
                url = reverse(self.OBJECT_PERM_URL, args=[user1.pk])
                self.assertTrue(f'href: "{url}"' in output)

    def test_with_arg(self):
        user1 = self.get_privileged_user()
        user2 = self.get_unprivileged_user()

        with self.setup_overrides():
            tpl = Template(
                """
            {% load react %}
            {% load alliance_ui %}
            {% component "a" href=perm|url_with_perm:user.pk|with_arg:2|with_arg:"abc123"|with_perm_obj:user %}{% endcomponent %}
            """
            )

            request = HttpRequest()
            request.user = user2
            request.session = SessionBase()
            context = Context({"request": request, "user": user1, "perm": self.MULTIPLE_ARGS_URL})
            output = tpl.render(context)
            self.assertEqual(output.strip(), "")
            request.user = user1
            output = tpl.render(context)
            url = reverse(self.MULTIPLE_ARGS_URL, args=[user1.pk, 2, "abc123"])
            self.assertTrue(f'href: "{url}"' in output)

    def test_url_no_perm_check(self):
        user1 = self.get_privileged_user()
        user2 = self.get_unprivileged_user()

        with self.setup_overrides():
            tpl = Template(
                """
            {% load react %}
            {% load alliance_ui %}
            {% component "a" href=perm|url:user.pk|with_arg:2|with_arg:"abc123" %}{% endcomponent %}
            """
            )

            request = HttpRequest()
            request.user = user2
            request.session = SessionBase()
            context = Context({"request": request, "user": user1, "perm": self.MULTIPLE_ARGS_URL})
            output = tpl.render(context)
            url = reverse(self.MULTIPLE_ARGS_URL, args=[user1.pk, 2, "abc123"])
            self.assertTrue(f'href: "{url}"' in output)



from pathlib import Path
import subprocess

from alliance_platform.frontend.bundler.asset_registry import FrontendAssetRegistry
from alliance_platform.frontend.bundler.vite import ViteBundler
from alliance_platform.frontend.settings import ap_frontend_settings
from bs4 import BeautifulSoup
from django.conf import settings

fixtures_dir = Path(__file__).parent.parent / "fixtures"

bundler_kwargs = dict(
    root_dir=settings.PROJECT_DIR,
    path_resolvers=[],
    build_dir=fixtures_dir / "build_test",
    server_build_dir=fixtures_dir / "server_build_test",
    server_host="localhost",
    server_port="5273",
    server_protocol="http",
    server_resolve_package_url="redirect-package-url",
)


def run_prettier(code):
    p = subprocess.run(
        [
            str(ap_frontend_settings.NODE_MODULES_DIR / ".bin/prettier"),
            "--stdin-filepath",
            "test.tsx",
        ],
        input=code,
        capture_output=True,
        text=True,
    )
    if p.returncode != 0:
        raise ValueError(f"Failed to format code: {p.stderr}")
    return p.stdout


def format_code(code: str):
    # Format the contents of the script tag using the dev code formatter
    soup = BeautifulSoup(code, "html.parser")
    script_tag = soup.find("script")
    # extract the contents of the script tag
    script_contents = script_tag.string  # type: ignore[union-attr]
    new_script_tag = soup.new_tag("script")
    new_script_tag.string = run_prettier(script_contents)

    # replace the old script tag with the new one
    script_tag.replace_with(new_script_tag)  # type: ignore[union-attr]
    return str(soup)


class TestViteBundler(ViteBundler):
    def validate_path(
        self,
        filename: str | Path,
        suffix_whitelist: list[str] | None = None,
        suffix_hint: str | None = None,
        resolve_extensions: list[str] | None = None,
    ) -> Path:
        return Path(filename)

    def format_code(self, code: str):
        return code


class TestFrontendAssetRegistryByPass(FrontendAssetRegistry):
    """Bypasses unknown checks by never returning any unknown paths"""

    def get_unknown(self, *filenames: Path) -> list[Path]:
        return []

    def add_asset(self, *filenames: Path):
        for filename in filenames:
            self._assets.add(filename)


bypass_frontend_asset_registry = TestFrontendAssetRegistryByPass()



from __future__ import annotations

from typing import Any

from allianceutils.auth.permission import reverse_if_probably_allowed
from django import template
from django.db.models import Model
from django.template import Context
from django.urls import reverse

from ..alliance_ui.button import register_button
from ..alliance_ui.date_picker import register_date_picker
from ..alliance_ui.icon import register_icon
from ..alliance_ui.inline_alert import register_inline_alert
from ..alliance_ui.menubar import register_menubar
from ..alliance_ui.misc import register_misc
from ..alliance_ui.pagination import register_pagination
from ..alliance_ui.table import register_table
from ..alliance_ui.time_input import register_time_input
from .react import DeferredProp
from .react import OmitComponentFromRendering

register = template.Library()

register_menubar(register)
register_table(register)
register_pagination(register)
register_button(register)
register_misc(register)
register_inline_alert(register)
register_icon(register)
register_date_picker(register)
register_time_input(register)


class NamedUrlDeferredProp(DeferredProp):
    """Used by ``url_with_perm`` and ``url`` filters to defer the resolution of a URL until rendering time.

    This allows perm checks to occur and to omit the component from rendering if failed"""

    url_name: str
    check_perm: bool
    args: list[Any]
    kwargs: dict[str, Any]
    object: Model | None

    def __init__(self, url_name, check_perm=False):
        self.url_name = url_name
        self.check_perm = check_perm
        self.args = []
        self.kwargs = {}
        self.object = None
        super().__init__()

    def set_object(self, arg: Model):
        self.object = arg

    def add_arg(self, arg: Any):
        self.args.append(arg)

    def add_kwargs(self, kwargs: dict[str, Any]):
        self.kwargs.update(kwargs)

    def resolve(self, context: Context):
        if self.check_perm:
            url = reverse_if_probably_allowed(
                context["request"], self.url_name, object=self.object, args=self.args, kwargs=self.kwargs
            )
            if url is None:
                raise OmitComponentFromRendering()
            return url
        return reverse(self.url_name, args=self.args, kwargs=self.kwargs)


@register.filter("url_with_perm")
def url_with_perm_filter(value, arg1=None):
    """Resolve a URL and check if the current user has permission to access it.

    If permission check fails, the component that uses the value will be omitted from rendering.

    Usage::

        {% component "a" href="my_url_name"|url_with_perm:2 %}

    The above example will resolve the URL "my_url_name" with the argument ``2``.

    If you need multiple arguments you can use the ``with_arg`` filter::


        {% component "a" href="my_url_name"|url_with_perm:2|with_arg:"another arg" %}

    To pass kwargs you can use the ``with_kwargs`` filter::

        {% component "a" href="my_url_name"|url_with_perm|with_kwargs:my_kwargs %}

    Note that as there's no way to define a dictionary in the standard Django template language, you'll need to
    pass it in context.

    To do object level permission checks use the ``with_perm_obj`` filter to pass through the object::

        {% component "a" href="my_url_name"|url_with_perm:obj.pk|with_obj:obj %}

    Note that you still have to pass through the pk to resolve the URL with. Passing the object just allows the permission
    checks to work without having to manually look up the object. Due to how the ``DetailView.get_object`` method works,
    if you are not going to pass the object you must use ``with_kwargs`` to pass through the ID rather than a positional
    argument::

        {% component "a" href="my_url_name"|url_with_perm|with_kwargs:kwargs %}

    Note that the above would do a query to retrieve the object, so it's better to pass the object if you have already
    retrieved it.
    """
    pfv = NamedUrlDeferredProp(value, True)
    if arg1 is not None:
        pfv.add_arg(arg1)
    return pfv


@register.filter("url")
def url_filter(value, arg1=None):
    """Behaves same as ``url_with_perm`` but does not check any permissions."""
    pfv = NamedUrlDeferredProp(value)
    if arg1 is not None:
        pfv.add_arg(arg1)
    return pfv


@register.filter("with_arg")
def with_arg(value: NamedUrlDeferredProp, arg):
    """Add an argument to a ``url_with_perm`` or ``url`` filter."""
    value.add_arg(arg)
    return value


@register.filter("with_kwargs")
def with_kwargs(value: NamedUrlDeferredProp, kwargs):
    """Add kwargs to a ``url_with_perm`` or ``url`` filter."""
    value.add_kwargs(kwargs)
    return value


@register.filter("with_perm_obj")
def with_perm_object(value: NamedUrlDeferredProp, arg):
    """Add an object to a ``url_with_perm`` filter for the purposes of object level permission checks."""
    if not isinstance(value, NamedUrlDeferredProp):
        raise ValueError(
            '`perm_obj` must appear after `url_with_perm`, eg. `"url_name"|url_with_perm:obj.pk|with_perm_obj:obj'
        )
    value.set_object(arg)
    return value


@register.filter("unwrap_list")
def unwrap_list(value):
    if isinstance(value, (list, tuple)):
        if not value:
            return None
        if len(value) != 1:
            raise ValueError(f"Expected list of length 1, received length {len(value)}: {value}")
        return value[0]
    return value



from functools import lru_cache
import json
from json import JSONDecodeError
import logging
from pathlib import Path
import re
import warnings

from alliance_platform.core.settings import ap_core_settings
from allianceutils.middleware import CurrentRequestMiddleware
from django.http import HttpRequest
import requests

from ..settings import ap_frontend_settings
from . import get_bundler
from .base import BaseBundler

logger = logging.getLogger("alliance_platform.frontend")


def resolve_vanilla_extract_cache_names(bundler: BaseBundler, filename: Path | str):
    cache_name = re.sub(r"[/.]", "_", str(filename))
    if bundler.is_development():
        cache_fn = ap_core_settings.CACHE_DIR / f"development-css-mappings/{cache_name}.json"
        import_script_filename = ap_core_settings.CACHE_DIR / f"development-css-mappings/{cache_name}.ts"
        return cache_fn, import_script_filename
    return ap_frontend_settings.PRODUCTION_DIR / "production-css-mappings" / f"{cache_name}.json", None


class VanillaExtractClassMapping:
    """Class that stores class mappings for a Vanilla Extract file

    The mappings are read from a JSON file written by the plugin defined in ``vanillaExtractWithExtras.ts``.

    If a class is referenced but does not exist a warning will be logged.

    In dev this triggers a request to the bundler to write the mappings if they don't exist. It also stores the import script
    filename so that it can be included in the template via ViteCssEmbed
    """

    #: Mapping of class names to their hashed names
    mapping: dict[str, str] | None
    bundler: BaseBundler
    cache_filename: Path

    _last_request: HttpRequest | None

    def __init__(self, bundler: BaseBundler, filename: Path):
        self._last_request = None
        self.filename = filename.relative_to(bundler.root_dir)
        self.bundler = bundler
        self.cache_filename, self._import_script_filename = resolve_vanilla_extract_cache_names(
            bundler, self.filename
        )
        self._create_mapping()

    @property
    def import_script_filename(self):
        """Only set in dev mode. This is a typescript file generated by ``vanillaExtractWithExtras.ts`` that imports the css and setups up HMR"""
        self._check_mapping()
        return self._import_script_filename

    def _create_mapping(self):
        bundler = self.bundler
        cache_fn = self.cache_filename
        if bundler.is_development():
            try:
                # Always trigger request otherwise we can end up with out of date mappings. It seems fast enough in practice.
                # Previously this checked if cache file exists and only triggered request if it didn't, but sometimes resulted
                # in out of date mappings.
                # This request will cause the above mappings file to be written out
                response = requests.get(
                    get_bundler().get_url(str(self.filename) + "?writeStyleMappings=1"), timeout=3
                )
                if response.status_code == 404:
                    warnings.warn(f"Requested stylesheet '{self.filename}' doesn't exist")
                elif response.status_code == 500:
                    warnings.warn(
                        f"Requested stylesheet '{self.filename}' could not be processed. Check for syntax errors and refresh."
                    )
                elif response.status_code == 200:
                    if not cache_fn.exists():
                        # In my testing this only happened if you removed the frontend cache dir after server was started.
                        warnings.warn(
                            f"Requested stylesheet '{self.filename}' exists but class name mappings could not be extracted. Try making a change to the file or restarting the dev server."
                        )
            except requests.exceptions.ConnectionError:
                warnings.warn(f"Failed to get requested stylesheet '{self.filename}'. Is 'yarn dev' running?")
            except requests.exceptions.Timeout:
                warnings.warn(f"Failed to get requested stylesheet '{self.filename}' - request timed out")
        if not cache_fn.exists():
            warnings.warn(
                f"Mapping for {self.filename} was not found in expected location '{cache_fn}'. No class names will be resolved for this tag."
            )
            self.mapping = None
        else:
            self._load_mapping()

    def _load_mapping(self):
        bundler = self.bundler
        cache_fn = self.cache_filename
        try:
            self.mapping = json.loads(cache_fn.read_text())
        except JSONDecodeError:
            if bundler.is_development():
                warnings.warn(
                    f"Failed to parse cache file {cache_fn}. This may be due to reading the file as it's being written to - try refreshing the browser."
                )
            else:
                logger.exception(f"Failed to parse cache file {cache_fn} - check file is valid JSON.")

    def _check_mapping(self):
        """If dev server has been restarted the cache dir will have been cleared.

        This detects such a case and regenerates files."""
        if self.bundler.is_development() and (
            not self.cache_filename.exists() or self.cache_filename.exists() and not self.mapping
        ):
            self._create_mapping()
        elif (
            self.bundler.is_development()
            and self.cache_filename.exists()
            and self._last_request != CurrentRequestMiddleware.get_request()
        ):
            # This handles the case where the cache file exists but may need to be reloaded. This can happen if the
            # class name mapping existed already, but details changed. For example, if you change:
            # style([class1, class2]) to style([class1]) then the key will exist in both cases, but the value will
            # change from something like '.class1 .class2' to '.class1'. Because of caching this won't be picked up otherwise.
            self._load_mapping()

        self._last_request = CurrentRequestMiddleware.get_request()

    def __getattr__(self, name):
        self._check_mapping()
        if self.mapping is None:
            warnings.warn(
                f"Requested class name '{name}' from '{self.filename}' cannot be resolved because the mapping file was not found."
            )
            return ""
        if name not in self.mapping:
            warnings.warn(
                f"Requested class name '{name}' from '{self.filename}' does not exist. Known classes are: {', '.join(self.mapping.keys())}"
            )
            return ""

        return self.mapping.get(name)

    def __str__(self):
        return f"VanillaExtractClassMapping({self.mapping})"


@lru_cache()
def resolve_vanilla_extract_class_mapping(bundler: BaseBundler, filename: Path):
    """Resolve the ``VanillaExtractClassMapping`` instance to use for the given filename"""
    return VanillaExtractClassMapping(bundler, filename)



from pathlib import Path

from django import template
from django.template import Library

from ..bundler.base import ResolveContext
from ..templatetags.react import ComponentNode
from ..templatetags.react import ImportComponentSource
from ..templatetags.react import parse_component_tag
from .utils import get_module_import_source


class PaginationNode(ComponentNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        resolver_context = ResolveContext(self.bundler.root_dir, self.origin.name)
        self._render_pagination_item_link_path = self.bundler.resolve_path(
            "@alliancesoftware/ui",
            resolver_context,
            resolve_extensions=[".ts", ".tsx", ".js"],
        )
        # This make the component use links so changing page navigates to new URL
        self.props["renderItem"] = ImportComponentSource(
            self._render_pagination_item_link_path, "renderPaginationItemAsLink", False
        )

    def get_paths_for_bundling(self) -> list[Path]:
        paths = super().get_paths_for_bundling()
        # Need to include this in the bundled paths so available in final build
        paths.append(self._render_pagination_item_link_path)
        return paths


def pagination(parser: template.base.Parser, token: template.base.Token):
    return parse_component_tag(
        parser,
        token,
        asset_source=get_module_import_source("@alliancesoftware/ui", "Pagination", False, parser.origin),
        node_class=PaginationNode,
    )


def register_pagination(register: Library):
    register.tag("Pagination")(pagination)



from os.path import dirname
from pathlib import Path
from unittest import mock

from alliance_platform.frontend.bundler.base import RegExAliasResolver
from alliance_platform.frontend.bundler.base import RelativePathResolver
from alliance_platform.frontend.bundler.base import ResolveContext
from alliance_platform.frontend.bundler.base import SourceDirResolver
from alliance_platform.frontend.bundler.base import html_target_browser
from alliance_platform.frontend.bundler.vite import ViteBundler
from django.conf import settings
from django.test import TestCase
from django.test import override_settings

from tests.test_utils import override_ap_frontend_settings


class TestBundler(TestCase):
    def test_relative_path_resolver(self):
        resolver = RelativePathResolver()
        # Can't resolve if no source_path specifier in context
        self.assertEqual(None, resolver.resolve("./test", ResolveContext(Path("/root"))))

        # Path.is_file will always return False if the file doesn't exist
        def fake_is_file(p):
            if str(p) == "/root/sub/template.html":
                return True
            return False

        with mock.patch("pathlib.Path.is_file", fake_is_file):
            self.assertEqual(
                Path("/root/sub/test"),
                resolver.resolve("./test", ResolveContext(Path("/root"), "/root/sub/template.html")),
            )
            self.assertEqual(
                Path("/root/test/myfile.js"),
                resolver.resolve(
                    "../test/myfile.js", ResolveContext(Path("/root"), "/root/sub/template.html")
                ),
            )

    def test_regex_alias_resolver(self):
        resolver = RegExAliasResolver(r"^node_modules/", "")
        # Can't resolve if no source_path specifier in context
        self.assertEqual(
            Path("package/test.ts"),
            resolver.resolve("node_modules/package/test.ts", ResolveContext(Path("/root"))),
        )

    def test_source_dir_resolver(self):
        resolver = SourceDirResolver(Path("/project/root/"))
        # Can't resolve if no source_path specifier in context
        self.assertEqual(
            Path("/project/root/package/test.ts"),
            resolver.resolve("package/test.ts", ResolveContext(Path("/root"))),
        )
        self.assertEqual(
            None,
            resolver.resolve("/project/test.ts", ResolveContext(Path("/root"))),
        )


class MockRequestResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


class TestViteBundler(ViteBundler):
    def validate_path(
        self,
        filename: str | Path,
        suffix_whitelist: list[str] | None = None,
        suffix_hint: str | None = None,
        resolve_extensions: list[str] | None = None,
    ) -> Path:
        """For test case don't validate the paths exist"""
        return Path(filename)


class TestViteBundlerTestCase(TestCase):
    def create_bundler(self, **kwargs):
        current_dir = Path(dirname(__file__))
        bundler_kwargs = dict(
            root_dir=settings.PROJECT_DIR,
            path_resolvers=[],
            build_dir=current_dir / "fixtures/build_test",
            server_build_dir=current_dir / "fixtures/server_build_test",
            mode="production",
            server_host="localhost",
            server_port="5273",
            server_protocol="http",
            server_resolve_package_url="redirect-package-url",
        )
        bundler_kwargs.update(kwargs)
        return TestViteBundler(**bundler_kwargs)  # type: ignore [arg-type]

    def test_dev_url(self):
        with self.assertRaisesMessage(ValueError, "static_url cannot be a full URL in dev"):
            with override_settings(STATIC_URL="http://mycdn.com/static/"):
                self.create_bundler(mode="development")

        with override_settings(STATIC_URL="/test-static/"):
            bundler = self.create_bundler(mode="development")
            self.assertEqual(
                "http://localhost:5273/test-static/components/Button.css",
                bundler.get_url(Path("components/Button.css")),
            )
            self.assertEqual(
                "http://localhost:5273/test-static/components/Button.tsx",
                bundler.get_url(Path("components/Button.tsx")),
            )

            # If absolute path is used should still work
            self.assertEqual(
                "http://localhost:5273/test-static/components/Button.tsx",
                bundler.get_url(settings.PROJECT_DIR / "components/Button.tsx"),
            )

            self.assertEqual("http://localhost:5273/ssr", bundler.get_development_ssr_url())

    @override_settings(STATIC_URL="/test-static/")
    def test_preview_url(self):
        bundler = self.create_bundler(mode="preview")
        self.assertEqual(
            "/test-static/assets/Button-abc123.css",
            bundler.get_url(Path("components/Button.css")),
        )
        self.assertEqual(
            "/test-static/assets/Button-def456.js",
            bundler.get_url(Path("components/Button.tsx")),
        )

        # If absolute path is used should still work
        self.assertEqual(
            "/test-static/assets/Button-def456.js",
            bundler.get_url(settings.PROJECT_DIR / "components/Button.tsx"),
        )

    @override_settings()
    def test_production_url(self):
        with override_settings(STATIC_URL="/test-static/"):
            bundler = self.create_bundler()
            self.assertEqual(
                "/test-static/assets/Button-abc123.css",
                bundler.get_url(Path("components/Button.css")),
            )
            self.assertEqual(
                "/test-static/assets/Button-def456.js",
                bundler.get_url(Path("components/Button.tsx")),
            )
        with override_settings(STATIC_URL="http://mycdn.com/static/"):
            cdn_bundler = self.create_bundler()
            self.assertEqual(
                "http://mycdn.com/static/assets/Button-abc123.css",
                cdn_bundler.get_url(Path("components/Button.css")),
            )
            self.assertEqual(
                "http://mycdn.com/static/assets/Button-def456.js",
                cdn_bundler.get_url(Path("components/Button.tsx")),
            )

            # If absolute path is used should still work
            self.assertEqual(
                "http://mycdn.com/static/assets/Button-def456.js",
                cdn_bundler.get_url(settings.PROJECT_DIR / "components/Button.tsx"),
            )

    def test_resolve(self):
        root_dir = settings.BASE_DIR
        bundler = self.create_bundler(
            root_dir=root_dir,
            path_resolvers=[
                RelativePathResolver(),
                RegExAliasResolver("^/", ""),
                SourceDirResolver(root_dir / "frontend/src"),
            ],
        )
        # The file needs to exist for is_file to work so hardcode path here - if this file is moved
        # then this test will break and will need to be manually updated
        context = ResolveContext(root_dir, "tests/test_bundler.py")
        self.assertEqual(
            root_dir / "another-app/file.js", bundler.resolve_path("/another-app/file.js", context)
        )

        self.assertEqual(
            root_dir / "tests/file.js",
            bundler.resolve_path("./file.js", context),
        )
        self.assertEqual(
            root_dir / "another-dir/file.js",
            bundler.resolve_path("../another-dir/file.js", context),
        )

    @override_settings(STATIC_URL="/test-static/")
    def test_collect_dependencies(self):
        bundler = self.create_bundler()
        deps = bundler.build_manifest.get_asset("components/TestComponent.tsx").collect_dependencies()
        self.assertEqual(
            deps.get_js_dependencies(),
            [
                "assets/TestComponent-47f28fe1.js",
                "assets/jsx-runtime-a1.js",
                "assets/TestComponentTwo-6347c19f.js",
                "assets/Button-def456.js",
            ],
        )
        self.assertEqual(
            deps.get_css_dependencies(),
            [
                "assets/TestComponent-ab23ac31.css",
                "assets/TestComponentTwo-b99abbde.css",
                "assets/Button-abc123.css",
            ],
        )
        self.assertEqual(deps.get_dynamic_js_dependencies(), ["assets/Input-b2.js"])
        self.assertEqual(deps.get_dynamic_css_dependencies(), ["assets/Input-c4.css"])

        deps2 = bundler.build_manifest.get_asset("components/TestComponentTwo.tsx").collect_dependencies()
        self.assertEqual(
            deps2.get_js_dependencies(),
            ["assets/TestComponentTwo-6347c19f.js", "assets/jsx-runtime-a1.js", "assets/Button-def456.js"],
        )
        deps2.merge(deps)
        self.assertEqual(
            deps2.get_js_dependencies(),
            [
                "assets/TestComponentTwo-6347c19f.js",
                "assets/jsx-runtime-a1.js",
                "assets/Button-def456.js",
                "assets/TestComponent-47f28fe1.js",
            ],
        )

    @override_settings(STATIC_URL="/test-static/")
    def test_embed(self):
        bundler = self.create_bundler()
        test_component_tags = [
            '<link rel="stylesheet" href="/test-static/assets/Button-abc123.css">',
            '<link rel="stylesheet" href="/test-static/assets/TestComponent-ab23ac31.css">',
            '<link rel="stylesheet" href="/test-static/assets/TestComponentTwo-b99abbde.css">',
            '<script src="/test-static/assets/TestComponent-47f28fe1.js" type="module"></script>',
        ]
        tags = [
            item.generate_code(html_target_browser)
            for item in bundler.get_embed_items("components/TestComponent.tsx")
        ]
        self.assertEqual(
            sorted(tags),
            test_component_tags,
        )
        # TestComponentTwo is a dependency of TestComponent so should have same set of tags
        # except for its javascript file. `get_embed_code` only includes the direct JS file for each
        # asset - it's dependencies are assumed to be load by the main asset script.
        self.assertEqual(
            sorted(
                [
                    item.generate_code(html_target_browser)
                    for item in bundler.get_embed_items(
                        [Path("components/TestComponent.tsx"), Path("components/TestComponentTwo.tsx")]
                    )
                ]
            ),
            test_component_tags
            + ['<script src="/test-static/assets/TestComponentTwo-6347c19f.js" type="module"></script>'],
        )

    @override_settings(STATIC_URL="/test-static/")
    def test_embed_standalone_css(self):
        bundler = self.create_bundler()
        items = bundler.get_embed_items("styles/normalize.css")
        self.assertEqual(len(items), 1)
        self.assertEqual(
            items[0].generate_code(html_target_browser),
            '<link rel="stylesheet" href="/test-static/assets/normalize-x1.css">',
        )

    def test_resolve_ssr_import_path(self):
        bundler = self.create_bundler()
        self.assertEqual(
            bundler.server_build_dir / "Button.mjs",
            bundler.resolve_ssr_import_path("components/Button.tsx"),
        )

    def test_format_code_size_limit(self):
        bundler = self.create_bundler(mode="development")

        def mocked_post(*args, **kwargs):
            return MockRequestResponse({"code": "formatted"}, 200)

        with mock.patch("requests.post", side_effect=mocked_post):
            self.assertEqual(bundler.format_code("a short string"), "formatted")
            long_string = "." * 1024 * 1024
            self.assertEqual(bundler.format_code(long_string), long_string)
            with override_ap_frontend_settings(DEV_CODE_FORMAT_LIMIT=0):
                self.assertEqual(bundler.format_code(long_string), "formatted")

    def test_format_code_timeout(self):
        bundler = self.create_bundler(mode="development")

        def mocked_post(*args, **kwargs):
            return MockRequestResponse({"code": "formatted"}, 200)

        with mock.patch("requests.post", side_effect=mocked_post) as mock_send:
            bundler.format_code("code")
            self.assertEqual(mock_send.call_args.kwargs.get("timeout"), 1)

        with mock.patch("requests.post", side_effect=mocked_post) as mock_send:
            with override_ap_frontend_settings(DEV_CODE_FORMAT_TIMEOUT=10):
                bundler.format_code("code")
                self.assertEqual(mock_send.call_args.kwargs.get("timeout"), 10)



from __future__ import annotations

from django import template
from django.template import Context
from django.template import Library

from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..templatetags.react import ComponentNode
from ..templatetags.react import ImportComponentSource
from ..templatetags.react import NestedComponentPropAccumulator
from ..templatetags.react import parse_component_tag


def fragment_component(parser: template.base.Parser, token: template.base.Token):
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
    source_path = get_bundler().resolve_path(
        "re-exports", resolver_context, resolve_extensions=[".ts", ".tsx", ".js"]
    )
    asset_source = ImportComponentSource(source_path, "Fragment", False)
    return parse_component_tag(parser, token, asset_source=asset_source)


class RawHtmlNode(ComponentNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.props["html"] = self.props.pop("children")

    def render(self, context: Context):
        accumulator = NestedComponentPropAccumulator.get_current(context)
        if accumulator:
            return super().render(context)
        return self.render_html(context)

    def render_html(self, context):
        return self.resolve_prop(self.props["html"], context)


def raw_html(parser: template.base.Parser, token: template.base.Token):
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
    source_path = bundler.resolve_path(
        "components/RawHtmlWrapper", resolver_context, resolve_extensions=[".ts", ".tsx", ".js"]
    )
    asset_source = ImportComponentSource(source_path, "RawHtmlWrapper", True)
    return parse_component_tag(parser, token, asset_source=asset_source, node_class=RawHtmlNode)


def register_misc(register: Library):
    register.tag("Fragment")(fragment_component)
    register.tag("raw_html")(raw_html)



from allianceutils.template import is_static_expression
from django import template
from django.template import Context
from django.template import Library
from django.template import TemplateSyntaxError
from django.template.base import Token

from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..templatetags.react import ImportComponentSource
from ..templatetags.react import parse_component_tag


def icon(parser: template.base.Parser, token: template.base.Token):
    """Render an icon from the core-ui/icons directory

    This tag accepts no children, so it requires no `endIcon`.

    Pass the name of the icon to the tag, and then any props you want to pass to the icon component.

    Usage::

       {% Icon "Pencil" data-testid="pencil" %}
    """
    contents = token.split_contents()
    icon_name_filter = parser.compile_filter(contents.pop(1))
    context = Context()
    if not is_static_expression(icon_name_filter):
        raise TemplateSyntaxError(
            f"Icon must be passed a static string for the icon name, got {icon_name_filter}"
        )

    icon_name = icon_name_filter.resolve(context)
    token = Token(token.token_type, " ".join(contents))
    bundler = get_bundler()
    origin = parser.origin
    resolver_context = ResolveContext(bundler.root_dir, origin.name if origin else None)
    # This was previous approach to avoid cascading requests for every icon included in the barrel file. This approach
    # proves to be unnecessary when using the Vite metadata to extract optimised deps - see ViteBundler.get_vite_dev_metadata
    # and it's usage in that file. The optimised file doesn't cause the cascade of requests.
    # sub_dir = "outlined"
    # if icon_name.endswith("Solid"):
    #     sub_dir = "solid"
    # elif icon_name.endswith("DuoTone"):
    #     sub_dir = "duotone"
    # elif icon_name.endswith("DuoColor"):
    #     sub_dir = "duocolor"
    # source_path = get_bundler().resolve_path(
    #     f"@alliancesoftware/icons/{sub_dir}/{icon_name}",
    #     resolver_context,
    #     resolve_extensions=[".ts", ".tsx", ".js"],
    # )
    source_path = get_bundler().resolve_path(
        "@alliancesoftware/icons",
        resolver_context,
        resolve_extensions=[".ts", ".tsx", ".js"],
    )
    asset_source = ImportComponentSource(source_path, icon_name, False)
    return parse_component_tag(parser, token, asset_source=asset_source, no_end_tag=True)


def register_icon(register: Library):
    register.tag("Icon")(icon)



from __future__ import annotations

import datetime
import math
from typing import TYPE_CHECKING
from typing import Any

from alliance_platform.codegen.typescript import ArrayLiteralExpression
from alliance_platform.codegen.typescript import Identifier
from alliance_platform.codegen.typescript import ImportSpecifier
from alliance_platform.codegen.typescript import NewExpression
from alliance_platform.codegen.typescript import convert_to_node
from django.template import Context
from django.utils.timezone import is_aware

from .bundler.ssr import SSRCustomFormatSerializable
from .bundler.ssr import SSRSerializerContext

if TYPE_CHECKING:
    from .templatetags.react import ComponentNode
    from .templatetags.react import ComponentSourceCodeGenerator


class CodeGeneratorNode:
    def generate_code(self, generator: ComponentSourceCodeGenerator):
        raise NotImplementedError()

    def convert_unknown(self, value, generator: ComponentSourceCodeGenerator):
        if isinstance(value, CodeGeneratorNode):
            return value.generate_code(generator)

    def convert_to_node(self, value, generator: ComponentSourceCodeGenerator):
        return convert_to_node(value, convert_unknown=lambda value: self.convert_unknown(value, generator))


class ComponentProp(SSRCustomFormatSerializable, CodeGeneratorNode):
    """
    Represents a prop that requires special conversion to generate the code and perform server side rendering.

    Simple props like strings, numbers, lists and dicts can all be converted automatically to the corresponding
    types in javascript. This works automatically as a) the code generator knows how to convert these types and
    b) they are JSON serializable so can be sent to the SSR server without any extra processing.

    Anything more complex than this needs to be handled by a custom prop handler. There are two parts to this:

    1) Implement ``generate_code`` to generate the typescript code that will be embedded in a <script> tag when
       ``ComponentNode`` is rendered.

    2) Implement ``get_tag`` and ``get_representation`` to return a JSON serializable representation of the prop that
       can then be revived on the frontend.

    A concrete example is ``DateProp``. This will generate code like::

        new Date("2020-01-01")

    such that the prop passed to the component is a ``Date`` instance. The ``get_representation`` method will return
    the string representation `"2020-01-01"` and ``get_tag`` returns ``"Date"``. This gets serialized to
    ``["@@CUSTOM", "Date", "2020-01-01"]``. The ``@@CUSTOM`` tag is used to indicate that the prop needs special "reviving"
    on the frontend which is detected by ``processSSRRequest`` in ``ssr.ts``, and will call the appropriate "reviver"
    defined in ``ssrJsonRevivers.tsx``.
    """

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        """Return ``True`` if this prop handler should be used for the given value."""
        raise NotImplementedError(f"should_apply not implemented for {cls.__name__}")

    def __init__(self, value: Any, node: ComponentNode, context: Context):
        """Intentionally blank; each implementation can handle this differently."""
        pass


class DateProp(ComponentProp):
    """Convert a python date or datetime to a js Date"""

    value: datetime.date

    def __init__(self, value, *args, **kwargs):
        super().__init__(value, *args, **kwargs)
        self.value = value
        self.js_args = [self.value.year, self.value.month, self.value.day]

    def get_tag(self):
        return "Date"

    def get_representation(self, context):
        return self.js_args

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        calendar_date = generator.resolve_prop_import(
            "frontend/src/re-exports.tsx", ImportSpecifier("CalendarDate")
        )
        return NewExpression(calendar_date, self.js_args)

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        return isinstance(value, datetime.date) and not isinstance(value, datetime.datetime)


class DateTimeProp(ComponentProp):
    """Convert a python datetime"""

    value: datetime.datetime

    def __init__(self, value, *args, **kwargs):
        super().__init__(value, *args, **kwargs)
        self.value = value
        self.js_args = [
            self.value.year,
            self.value.month,
            self.value.day,
            self.value.hour,
            self.value.minute,
            self.value.second,
            self.value.microsecond,
        ]

    def get_tag(self):
        return "DateTime"

    def get_representation(self, context):
        return self.js_args

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        calendar_date = generator.resolve_prop_import(
            "frontend/src/re-exports.tsx", ImportSpecifier("CalendarDateTime")
        )
        return NewExpression(
            calendar_date,
            self.js_args,
        )

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        return isinstance(value, datetime.datetime) and not is_aware(value)


class ZonedDateTimeProp(ComponentProp):
    """Convert a python datetime"""

    value: datetime.datetime
    js_args: list[int | str]

    def __init__(self, value, *args, **kwargs):
        super().__init__(value, *args, **kwargs)
        offset = value.utcoffset()
        if offset is None:
            raise ValueError("ZonedDateTimeProp should only be used with aware datetimes")
        self.value = value
        self.js_args = [
            self.value.year,
            self.value.month,
            self.value.day,
            # e.g. Australia/Melbourne
            str(self.value.tzinfo),
            int(offset.total_seconds() * 1000),
            self.value.hour,
            self.value.minute,
            self.value.second,
            self.value.microsecond,
        ]

    def get_tag(self):
        return "ZonedDateTime"

    def get_representation(self, context):
        return self.js_args

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        calendar_date = generator.resolve_prop_import(
            "frontend/src/re-exports.tsx", ImportSpecifier("ZonedDateTime")
        )
        return NewExpression(
            calendar_date,
            self.js_args,
        )

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        return isinstance(value, datetime.datetime) and is_aware(value)


class TimeProp(ComponentProp):
    """Convert a python time to a @internationalized/date time"""

    value: datetime.time

    def __init__(self, value, *args, **kwargs):
        super().__init__(value, *args, **kwargs)
        self.value = value
        self.js_args = [self.value.hour, self.value.minute, self.value.second, self.value.microsecond]

    def get_tag(self):
        return "Time"

    def get_representation(self, context):
        return self.js_args

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        calendar_date = generator.resolve_prop_import("frontend/src/re-exports.tsx", ImportSpecifier("Time"))
        return NewExpression(calendar_date, self.js_args)

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        return isinstance(value, datetime.time)


class SetProp(ComponentProp):
    """Handles passing a python ``set`` to a JS ``Set``"""

    value: set

    def get_tag(self):
        return "Set"

    def get_representation(self, context: SSRSerializerContext) -> list:
        return list(self.value)

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        return NewExpression(
            Identifier("Set"),
            [ArrayLiteralExpression([self.convert_to_node(value, generator) for value in self.value])],
        )

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        return isinstance(value, set)

    def __init__(self, value: set, node: ComponentNode, context: Context):
        super().__init__(value, node, context)
        self.value = set({node.resolve_prop(element, context) for element in value})


class SpecialNumeric(ComponentProp):
    """Special prop that indicates a prop should be passed as NaN or Infinity/-Infinity

    This is used whenever a math.inf or math.nan / float("nan") is passed as a prop.
    """

    value: float

    def __init__(self, value: float, node: ComponentNode, context: Context):
        super().__init__(value, node, context)
        self.value = value

    def get_tag(self):
        return "SpecialNumeric"

    def generate_code(self, generator: ComponentSourceCodeGenerator):
        if math.isinf(self.value):
            return Identifier("Infinity" if self.value > 0 else "-Infinity")
        return Identifier("NaN")

    def get_representation(self, context: SSRSerializerContext) -> dict | str | list:
        if math.isinf(self.value):
            return "Infinity" if self.value > 0 else "-Infinity"
        return "NaN"

    @classmethod
    def should_apply(cls, value: Any, node: ComponentNode, context: Context):
        return isinstance(value, float) and (math.isinf(value) or math.isnan(value))


default_prop_handlers = [DateProp, DateTimeProp, ZonedDateTimeProp, TimeProp, SetProp, SpecialNumeric]



from pathlib import Path
import warnings

from allianceutils.template import is_static_expression
from allianceutils.template import parse_tag_arguments
from allianceutils.template import resolve
from django import template
from django.template import Context
from django.template import Origin
from django.template.base import UNKNOWN_SOURCE
from django.template.base import FilterExpression
from django.template.loader import get_template
from django.utils.safestring import mark_safe

from ..bundler import get_bundler
from ..bundler.base import ResolveContext
from ..bundler.base import html_target_browser
from ..bundler.context import BundlerAsset
from ..bundler.context import BundlerAssetContext
from ..settings import ap_frontend_settings

register = template.Library()


class BundlerUrlAssetNode(template.Node, BundlerAsset):
    """
    The node used for the ``bundler_url`` tag.
    """

    def __init__(self, path: Path, origin: Origin | None, target_var: str | None = None):
        super().__init__(origin or Origin(UNKNOWN_SOURCE))
        self.path = path
        self.target_var = target_var

    def get_paths_for_bundling(self):
        return [self.path]

    def render(self, context):
        url = self.bundler.get_url(self.path)
        if self.target_var:
            context[self.target_var] = url
            return ""
        return url


class BundlerEmbedAssetNode(template.Node, BundlerAsset):
    """
    The node used for the ``bundler_embed`` tag
    """

    def __init__(
        self,
        path: Path,
        origin: Origin | None,
        target_var: str | None = None,
        inline=False,
        content_type: str | None = None,
        html_attrs: dict[str, str] | None = None,
    ):
        self.path = path
        self.target_var = target_var
        self.inline = inline
        self.html_attrs = html_attrs or {}
        self.content_type = content_type

        super().__init__(origin or Origin(UNKNOWN_SOURCE))

    def get_paths_for_bundling(self):
        return [self.path]

    def render(self, context):
        items = self.bundler.get_embed_items(self.get_paths_for_bundling(), self.content_type)
        tags = []
        for item in items:
            item.html_attrs = resolve(self.html_attrs, context)
            if not self.inline and not self.target_var and item.can_embed_head():
                self.bundler_asset_context.queue_embed_file(item)
            else:
                tags.append(item.generate_code(html_target_browser))
        code = mark_safe("\n".join(tags))
        if self.target_var:
            context[self.target_var] = code
            return ""
        return code


@register.tag("bundler_url")
def bundler_url(parser: template.base.Parser, token: template.base.Token):
    """
    Return the URL from the bundler to a specified asset.

    NOTE: This tag only accepts static values as :class:`extract_frontend_assets <alliance_platform.frontend.management.commands.extract_frontend_assets.Command>` needs to be able to statically generate
    a list of paths that need to be built for production.

    Usage::

        {% load bundler %}

        {% bundler_url "view.css" %}
    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)

    if len(args) != 1:
        raise template.TemplateSyntaxError(
            f"{tag_name} must receive a single argument which is the path to the file to include"
        )

    if not is_static_expression(args[0]):
        raise template.TemplateSyntaxError(
            f"{tag_name} must be passed a static string for it's argument (encountered variable '{args[0].var}')"
        )
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
    filename = bundler.resolve_path(args[0].resolve(Context()), resolver_context)

    return BundlerUrlAssetNode(filename, parser.origin, target_var=target_var)


@register.simple_tag()
def bundler_preamble():
    """Adds necessary code for things like enabling HMR.

    Typically this is only required in development but that is up to the Bundler to decide - the tag should
    be included for both production and development.

    Usage::

        {% load bundler %}

        // In the <head> element
        {% bundler_preamble %}
    """
    return mark_safe(get_bundler().get_preamble_html())


@register.tag("bundler_embed")
def bundler_embed(parser: template.base.Parser, token: template.base.Token):
    """
    Return the embed HTML codes from the bundler to a specified asset.

    Each asset can have multiple files associated with it. For example, a component might have javascript and css. You
    can control which types of tags are included using the ``content_type`` kwarg. Common types are ``text/css`` and ``text/javascript``,
    but it is ultimately based on the file extension (e.g. ``.png`` will be ``image/png``). Note that ``.css.ts`` is handled
    as ``text/css`` and ``.ts`` and ``.tsx`` are handled as ``text/javascript``.

    By default, the tags are added to the HTML by the :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_embed_collected_assets`.
    This allows assets to be embedded as needed in templates but all added in one place in the HTML (most likely the ``<head>``).
    You can force the tags to be outputted inline with ``inline=True``.

    NOTE: This tag only accepts static values as :class:`extract_frontend_assets <alliance_platform.frontend.management.commands.extract_frontend_assets.Command>`
    needs to be able to statically generate a list of paths that need to be built for production.

    Args:
        content_type: (optional) If set to either 'css' or 'js' only assets of the matching type will be embedded. If omitted
            both types will be included (if available).
        inline: If ``True`` the tags will be embedded inline, otherwise they will be added using the
            :meth:`~alliance_platform.frontend.templatetags.bundler.bundler_embed_collected_assets` tag. Defaults to ``False``.
        html_*: any parameter with a ``html_`` prefix will have the ``html_`` stripped and passed through to the embed tag
            eg ``html_id="foo"`` would render ``<script id="foo" ...>`` for script tags

    Usage with ``bundler_embed_collected_assets``:

    .. code-block:: html

        // in the base template (e.g. base.html)
        <!doctype html>
        {% load bundler %}
        <html lang="en-AU">
          <head>
            {% bundler_embed_collected_assets %}
          </head>
          <body>{% block body %}{% endblock %}</body>
        </html>

        // in other individual templates, e.g. 'myview.html'

        {% extends "base.html" %}
        {% block body %}
            {% bundler_embed "MyComponent.ts" %}
            {% bundler_embed "logo.png" html_alt="My Component Logo" %}
            <h1>My View</h1>
        {% endblock %}

    would output:

    .. code-block:: html

        <!doctype html>
        <html lang="en-AU">
          <head>
            <script type="module" src="http://localhost:5173/assets/MyComponent.js"></script>
            <link rel="stylesheet" href="http://localhost:5173/assets/MyComponent.css" />
          </head>
          <body>
            <img src="http://localhost:5173/assets/logo.png" alt="My Component Logo" />
            <h1>My View</h1>
          </body>
        </html>

    Using ``inline=True`` instead:

    .. code-block:: html

        {% extends "base.html" %}
        {% block body %}
            {% bundler_embed "MyComponent.ts" inline=True %}
            <h1>My View</h1>
        {% endblock %}

    would output:

    .. code-block:: html

        <!doctype html>
        <html lang="en-AU">
          <head></head>
          <body>
            <script type="module" src="http://localhost:5173/assets/MyComponent.js"></script>
            <link rel="stylesheet" href="http://localhost:5173/assets/MyComponent.css" />
            <h1>My View</h1>
          </body>
        </html>

    Note that in the example above ``logo.png`` is always embedded inline as it is not a javascript or css file.
    """
    tag_name = token.split_contents()[0]
    args, kwargs, target_var = parse_tag_arguments(parser, token, supports_as=True)
    if len(args) != 1:
        raise template.TemplateSyntaxError(
            f"{tag_name} must receive a single argument which is the path to the file to include"
        )

    if not is_static_expression(args[0]):
        raise template.TemplateSyntaxError(
            f"{tag_name} must be passed a static string for it's argument (encountered variable '{args[0].var}')"
        )
    content_type: FilterExpression | None = kwargs.pop("content_type", None)
    if content_type is not None and not is_static_expression(content_type):
        raise template.TemplateSyntaxError(
            f"{tag_name} must be passed a static string ('css' or 'js') for 'content_type' (encountered variable '{content_type.var}')"
        )
    inline: FilterExpression | None = kwargs.pop("inline", None)
    if inline is not None and not is_static_expression(inline):
        raise template.TemplateSyntaxError(
            f"{tag_name} must be passed a static value of `True` or `False` for 'inline' (encountered variable '{inline.var}')"
        )
    html_attrs = {}
    for k, v in list(kwargs.items()):
        if k.startswith("html_"):
            html_attrs[k[5:]] = v
            kwargs.pop(k)
    if kwargs:
        raise template.TemplateSyntaxError(f"{tag_name} received unknown kwargs: {', '.join(kwargs.keys())}")
    # context used to resolve static expressions
    static_context = Context()
    content_type = content_type.resolve(static_context) if content_type else None
    bundler = get_bundler()
    resolver_context = ResolveContext(bundler.root_dir, parser.origin.name if parser.origin else None)
    filename = bundler.resolve_path(args[0].resolve(static_context), resolver_context)
    return BundlerEmbedAssetNode(
        filename,
        parser.origin,
        target_var=target_var,
        content_type=content_type,
        inline=inline,
        html_attrs=html_attrs,
    )


@register.simple_tag()
def bundler_embed_collected_assets():
    """Add tags to header for assets required in page

    This works with :class:`~alliance_platform.frontend.bundler.context.BundlerAssetContext` to collect all the assets used
    within a template. See :class:`~alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware` for how
    this context is created for you in Django views.

    Because each asset must specify asset paths statically, this tag can retrieve assets from ``BundlerAssetContext``
    and embed the required tags before the rest of the template is rendered.

    Some existing assets are those created by the :func:`~alliance_platform.frontend.templatetags.vanilla_extract.stylesheet`,
    :func:`~alliance_platform.frontend.templatetags.react.component`, or :func:`~alliance_platform.frontend.templatetags.bundler.bundler_embed`
    tags. See the individual implementations for options that may influence how they are embedded (e.g. the ``inline``
    option provided by ``bundler_embed``).

    :data:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.html_target` will control whether scripts are included
    and whether CSS is outputted in line in ``style`` tags or linked externally.

    Generally, this tag should be used in the ``<head>`` of the HTML document. All script tags are non-blocking by default.
    """
    asset_context = BundlerAssetContext.get_current()
    return asset_context.register_embed_collected_assets_tag()


@register.simple_tag()
def bundler_dev_checks():
    """Performs dev specific checks and may render some HTML to communicate messages to user

    Currently check if the dev server is running for this project, and if not displays an error.

    Error will be logged to Django dev console. In addition, an error icon and toggleable modal message will be shown
    in the HTML unless :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.BUNDLER_DISABLE_DEV_CHECK_HTML` is set.
    """
    bundler = get_bundler()
    if not bundler.is_development():
        return ""
    check = bundler.check_dev_server()
    if not check.is_ok():
        if not check.is_running:
            warnings.warn("Bundler dev server not running; run `yarn dev` to start")
        elif check.is_wrong_server():
            warnings.warn(
                f"Bundler dev server was found but it's for project {check.project_dir}. You likely need to run `yarn dev` under {bundler.root_dir}"
            )
        elif check.read_timeout:
            warnings.warn(
                "Received ReadTimeout from dev server check; dev server is likely running but no response was received within timeout to verify"
            )
        if ap_frontend_settings.BUNDLER_DISABLE_DEV_CHECK_HTML:
            return ""
        return get_template("bundler_dev_check.html").render({"check": check})
    return ""



from pathlib import Path
import threading
from typing import Callable
from unittest import mock

from alliance_platform.frontend.bundler.context import BundlerAsset
from alliance_platform.frontend.bundler.context import BundlerAssetContext
from django.template import Origin
from django.template.base import UNKNOWN_SOURCE
from django.test import TestCase
from django.test import modify_settings
from django.test import override_settings
from django.urls import reverse


def execute_request(
    client: object,
    url_path: str,
    data: dict[str, str | int] | None = None,
    thread_count: int = 1,
    prehook: Callable[[object, int], object] | None = None,
):
    """
    Execute a request, optionally on multiple threads

    Taken from https://github.com/AllianceSoftware/django-allianceutils/blob/f31e9ef343ef6ade4776ebcbbb0c22445c5bc1f9/test_allianceutils/tests/middleware/tests.py#L22

    :param url_path: URL path to request
    :param data: POST variables
    :param thread_count: number of threads to create & run this request in
    :return: a list of responses if `thread_pool` more than 1, otherwise a single response
    """
    thread_exceptions = []
    thread_responses = []

    def do_request(client=None, count=None):
        try:
            if prehook:
                client = prehook(client, count)
            response = client.post(path=url_path, data=data, content_type="application/json")
            thread_responses.append(response)
        except Exception as ex:
            thread_exceptions.append(ex)
            raise

    if thread_count == 1:
        do_request(client, 0)
        return thread_responses[0]

    threads = [threading.Thread(target=do_request, args=(client, count)) for count in range(thread_count)]

    for t in threads:
        t.start()
    for t in threads:
        t.join()

    if thread_exceptions:
        raise Exception(f"Found {len(thread_exceptions)} exception(s): {thread_exceptions}")

    return thread_responses


@override_settings(ROOT_URLCONF="test_alliance_platform_frontend.urls")
@modify_settings(
    MIDDLEWARE={
        "remove": ["stronghold.middleware.LoginRequiredMiddleware"],
        "append": [
            "alliance_platform.frontend.bundler.middleware.BundlerAssetContextMiddleware",
        ],
    },
)
class BundlerAssetContextTestCase(TestCase):
    def setUp(self):
        self.path = reverse("bundler_container_ids")

    def test_container_ids_unique(self):
        """Make sure container ids are unique when multiple threads used"""
        thread_count = 13
        container_count = 3
        responses = execute_request(
            client=self.client,
            url_path=self.path,
            thread_count=thread_count,
            data={"containerCount": container_count},
        )
        container_ids = []
        for response in responses:
            container_ids += response.json()["container_ids"]
        self.assertEqual(len(container_ids), thread_count * container_count)
        self.assertEqual(len(container_ids), len(set(container_ids)))

    def test_middleware_queue_ssr_json_response(self):
        # Can't SSR if response isn't text/html - check for this
        with mock.patch(
            "alliance_platform.frontend.bundler.ssr.BundlerAssetServerSideRenderer.process_ssr"
        ) as mock_method:
            with self.assertWarns(
                UserWarning, msg="There are items to SSR but the response is application/json"
            ):
                self.client.post(
                    reverse("bundler_ssr"),
                    data={"items": ["item 1", "item 2"]},
                    content_type="application/json",
                )
                mock_method.assert_not_called()

    def test_middleware_queue_ssr(self):
        with mock.patch(
            "alliance_platform.frontend.bundler.ssr.BundlerAssetServerSideRenderer.process_ssr"
        ) as mock_method:
            # Mock server rendering to just return dummy html for each item
            mock_method.return_value = {
                "renderedItems": {
                    "<!-- ___SSR_PLACEHOLDER_0___ -->": {"html": "<rendered item 1>"},
                    "<!-- ___SSR_PLACEHOLDER_1___ -->": {"html": "<rendered item 2>"},
                },
                "errors": {},
            }
            response = self.client.post(
                reverse("bundler_ssr"),
                data={"items": ["item 1", "item 2"]},
                content_type="application/json",
                HTTP_ACCEPT="text/html",
            )
            self.assertEqual(
                [
                    "item 1: <rendered item 1>",
                    "item 2: <rendered item 2>",
                ],
                response.content.decode().split("\n"),
            )

    def test_asset_context_get_assets(self):
        my_asset_paths = [Path("/path1")]

        class MyAsset(BundlerAsset):
            def get_paths_for_bundling(self) -> list[Path]:
                return my_asset_paths

        my_other_asset_paths = [Path("/other/path1"), Path("/other/path2")]

        class MyOtherAsset(BundlerAsset):
            def get_paths_for_bundling(self) -> list[Path]:
                return my_other_asset_paths

        with BundlerAssetContext(skip_checks=True) as context:
            origin = Origin(UNKNOWN_SOURCE)
            asset1 = MyAsset(origin)
            asset2 = MyOtherAsset(origin)
            self.assertEqual(asset1.bundler_asset_context, context)
            self.assertEqual(set(context.get_assets()), {asset1, asset2})
            self.assertEqual(set(context.get_assets(MyAsset)), {asset1})
            self.assertEqual(set(context.get_asset_paths()), {*my_asset_paths, *my_other_asset_paths})
            self.assertEqual(set(context.get_asset_paths(MyOtherAsset)), set(my_other_asset_paths))



from contextlib import redirect_stdout
from functools import lru_cache
import io
import json
from pathlib import Path
import re
import sys
from typing import Any
from typing import Collection

from django.core.management import BaseCommand
from django.core.management.base import OutputWrapper
from django.template import TemplateSyntaxError
from django.template import engines
from django.template.loader import get_template
from django.template.utils import get_app_template_dirs

from ...bundler import get_bundler
from ...bundler.context import BundlerAssetContext
from ...settings import ap_frontend_settings


@lru_cache
def get_all_templates_files() -> list[Path]:
    """Scans all template dirs for template files

    Will exclude any templates the match an entry in :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.EXTRACT_ASSETS_EXCLUDE_DIRS`
    """
    dirs = get_app_template_dirs("templates")
    for engine in engines.all():
        dirs += tuple(engine.template_dirs)
    files: list[Path] = []
    for dir in dirs:
        should_exclude = False
        for excl in ap_frontend_settings.EXTRACT_ASSETS_EXCLUDE_DIRS:
            if isinstance(excl, re.Pattern):
                if excl.match(str(dir)):
                    should_exclude = True
                    break
            elif str(dir).startswith(str(excl)):
                should_exclude = True

        if should_exclude:
            continue
        files.extend(x for x in Path(dir).glob("**/*.html") if x)
    return files


def extract_asset_filenames() -> tuple[list[Any], dict[str, Collection[str]], list[str], list[str]]:
    """Scans all template files for assets that need to be bundled

    Returns a 4-tuple: list of filenames, breakdown dict, errors and warnings
    """

    all_files = {str(f) for f in ap_frontend_settings.FRONTEND_ASSET_REGISTRY.get_asset_paths()}
    breakdown_templates: dict[str, list[str]] = {}
    breakdown = {
        "registry": list(sorted(all_files)),
        "templates": breakdown_templates,
    }
    errors: list[str] = []
    warnings: list[str] = []
    with BundlerAssetContext() as asset_context:
        prev = asset_context.get_asset_paths()
        for file in get_all_templates_files():
            # TODO: ability to opt out specific templates?
            try:
                get_template(str(file))
                template_assets = [p for p in asset_context.get_asset_paths() if p not in prev]
                prev = asset_context.get_asset_paths()
                if template_assets:
                    breakdown_templates[str(file)] = [str(p) for p in sorted(template_assets)]
            except TemplateSyntaxError:
                warnings.append(
                    f"Failed to parse {file.relative_to(get_bundler().root_dir)} - any tags in that file will be ignored"
                )
        paths = [str(p) for p in asset_context.get_asset_paths()]
        all_files.update(paths)
        breakdown["templates"] = dict(sorted(breakdown_templates.items()))

    node_modules_dir = ap_frontend_settings.NODE_MODULES_DIR

    def transform_path(path: Path):
        if path.is_relative_to(node_modules_dir):
            return str(path.relative_to(node_modules_dir))
        return str(path)

    return [transform_path(Path(f)) for f in all_files], breakdown, errors, warnings


class Command(BaseCommand):
    """
    Extracts used frontend assets from templates used

    This works with any template nodes that extend :class:`~alliance_platform.frontend.bundler.context.BundlerAsset`. All templates
    in the system are loaded to gather all used assets. You can exclude specific directories by setting
    :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.EXTRACT_ASSETS_EXCLUDE_DIRS`
    to either a :class:`pathlib.Path` or ``re.Pattern``.

    Outputs a valid JSON dump as an array of string paths to files.

    Usage::

        # Will write data to output.json
        ./manage.py extract_frontend_assets --output output.json

        # Will write to stdin and supress other std output
        ./manage extract_frontend_assets --quiet
    """

    help = "Outputs JSON dump of files that need to be bundled for production build"

    def add_arguments(self, parser):
        parser.add_argument(
            "--quiet", action="store_true", help="If set any stdout output will be suppressed"
        )
        parser.add_argument(
            "--output",
            help="The path to write the output to. If not set output will be written to stdout",
        )

    def handle(self, *args, quiet=False, output=None, **kwargs):
        stderr_warn = OutputWrapper(sys.stderr)
        stderr_warn.style_func = self.style.WARNING
        f = io.StringIO()
        with redirect_stdout(f if quiet else self.stdout):  # type: ignore[type-var]
            files, breakdown, errors, warnings = extract_asset_filenames()
            if warnings:
                stderr_warn.write("\n".join(warnings))
            if errors:
                self.stderr.write("\n".join(errors))
            if errors:
                sys.exit(1)
            data = json.dumps({"files": files, "breakdown": breakdown})
            if output:
                Path(output).write_text(data)
            else:
                self.stdout.write(data)



from .base import BaseBundler


# Removed caching as it causes problems with tests is you don't force the cache clear (e.g. if one test
# sets FRONTEND_BUNDLER it's value will be cached for other tests - so order matters).
# Probably makes very little difference anyway - looks like `import_string` itself does some caching.
# @lru_cache
def get_bundler() -> BaseBundler:
    """Get the current bundler instance

    This comes from the  :data:`~alliance_platform.frontend.settings.AlliancePlatformFrontendSettingsType.BUNDLER` setting
    """
    from ..settings import ap_frontend_settings

    return ap_frontend_settings.BUNDLER



import logging
import warnings

from django.http import HttpRequest
from django.http import HttpResponse

from .context import BundlerAssetContext

logger = logging.getLogger("alliance_platform.frontend")


class BundlerAssetContextMiddleware:
    """Middleware that wraps requests in a :class:`~alliance_platform.frontend.bundler.context.BundlerAssetContext` manager

    Within a Django request you can call :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.get_current` and
    it will return context instance for the current request. :class:`~alliance_platform.frontend.bundler.context.BundlerAssetContext`
    does this on init.

    This is currently only used to facilitate :ref:`server side rendering <ssr>` by collecting all the items that need to
    be rendered. This works as follows:

    - :class:`~alliance_platform.frontend.bundler.context.SSRItem` can be queued for server side rendering using
      :meth:`~alliance_platform.frontend.bundler.context.BundlerAssetContext.queue_ssr`. The caller (e.g. a template node) must
      render a placeholder element in the page, e.g. ``<!-- ___SSR_PLACEHOLDER_0___ -->`` - this placeholder string is
      returned by ``queue_ssr``
    - The middleware processes the request like normal so all the HTML is generated, including all the necessary placeholders.
    - It then does the SSR for all requested items.
    - For each item, it will replace the placeholder in the HTML with the rendered version.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request: HttpRequest):
        with BundlerAssetContext(request=request) as asset_context:
            response: HttpResponse = self.get_response(request)
            if not asset_context.requires_post_processing():
                # No ssr requests - can return response directly
                return response
            if response.has_header("content-type"):
                # Warn if SSR has been requested but response isn't HTML
                if not response.headers["content-type"].startswith("text/html"):
                    asset_context.abort_post_process()
                    warnings.warn(
                        f"There are items to SSR but the response is {response['content-type']} - items have been ignored"
                    )
                    return response
            response.content = asset_context.post_process(response.content.decode()).encode()

            return response
